# -*- coding: utf-8 -*-

import collections
import datetime
import sqlalchemy as sa

from repoze.workflow import get_workflow
from pyramid.view import view_config

from penelope.core import fanstatic_resources
from penelope.core.lib.helpers import unicodelower, total_seconds
from penelope.core.models import DBSession
from penelope.core.models.dashboard import Project, CustomerRequest
from penelope.core.models.tickets import ticket_store


class ProjectBGB(object):
    """
    Backlog-Grooming-Board table for each project
    """
    def __init__(self, project, estimations_cache, done_cache):
        self.project = project
        self.estimations_cache = estimations_cache
        self.done_cache = done_cache

    def get_estimate(self, cr):
        td = self.estimations_cache.get(cr.id, datetime.timedelta(0))
        return total_seconds(td)

    def get_done(self, cr):
        td = self.done_cache.get(cr.id, datetime.timedelta(0))
        return total_seconds(td)

    def get_percentage(self, cr):
        estimate = self.get_estimate(cr)
        if not estimate:
            return 0
        return float(self.get_done(cr)) / estimate * 100


class Backlog(object):
    def __init__(self, request):
        self.request = request
        self.context = request.context


    def fetch_estimations(self, projects):
        project_ids = [a.id for a in projects]
        return dict(
                (row.id, datetime.timedelta(days=row.estimation_days/3.0))    # 8 vs 24 hour days
                    for row in DBSession().query(CustomerRequest)\
                                          .filter(CustomerRequest.project_id.in_(project_ids))
                )

    def fetch_done(self, projects):
        cr_done = collections.defaultdict(lambda: datetime.timedelta(0))
        for project in projects:
            for t in project.time_entries:
                cr_done[t.customer_request_id] += t.hours
        return cr_done

    @view_config(name='backlog', renderer='skin', permission='view_backlog', http_cache=0)
    def backlog(self, projects=None):
        fanstatic_resources.backlog.need()

        session = DBSession()

        if projects is None:
            projects = session.query(Project)
            projects = sorted(self.request.filter_viewables(projects.filter(Project.active)),
                              key=lambda p: (p.customer.name.lower(), p.customer.name.lower()))

        estimations_cache = self.fetch_estimations(projects)
        done_cache = self.fetch_done(projects)

        bgbs = [ProjectBGB(p, estimations_cache, done_cache) for p in projects]

        workflow = get_workflow(CustomerRequest(), 'CustomerRequest')

        cr_workflow_states = [
                    (state, workflow._state_data[state]['title'] or state)
                    for state in workflow._state_order
                ]

        #------------
        # Permissions
        #------------

        can_view_done = collections.defaultdict(bool)
        can_view_done_column = False

        can_view_percentage = collections.defaultdict(bool)
        can_view_percentage_column = False

        can_view_estimate = collections.defaultdict(bool)
        can_view_estimate_column = False
        can_edit_cr = collections.defaultdict(bool)

        for project in projects:

            if self.request.has_permission('estimations', project):
                # display values in the project's "estimate" column, and the project's total
                can_view_estimate[project] = True
                can_view_done[project] = True
                # display values in the project's "done" column, and the project's total
                can_view_percentage[project] = True
                # display a big-total and inserts an "estimate" column in the table
                can_view_estimate_column = True
                # display a big-total and inserts a "done" column in the table
                can_view_done_column = True
                can_view_percentage_column = True

            for cr in project.customer_requests:
                if self.request.has_permission('edit', cr):
                    # let the user change the CR
                    can_edit_cr[cr] = True

        return {
            'bgbs': bgbs,
            'unicodelower': unicodelower,
            'multiple_bgb': True,
            'cr_workflow_states': cr_workflow_states,
            'cr_workflow_active': set(['created', 'estimated', 'achieved', 'invoiced']),
            'can_view_percentage': can_view_percentage,
            'can_view_percentage_column': can_view_percentage_column,
            'can_view_done': can_view_done,
            'can_view_done_column': can_view_done_column,
            'can_view_estimate': can_view_estimate,
            'can_view_estimate_column': can_view_estimate_column,
            'can_edit_cr': can_edit_cr
            }




"this code is shared from ptahproject under BSD license"

import logging
import zope.component

from beaker.cache import cache_region
from beaker.cache import region_invalidate
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Allow, Deny
from pyramid.security import Authenticated
from zope.interface import Interface

from penelope.core.interfaces import IRoleFinder
from penelope.core.models.interfaces import IProjectRelated, ICustomer, IUser, IDublinCore


class ACL(list):
    """ Named ACL map

    ACL contains list of permit rules, for example::

      >> acl = ACL('test', 'Test ACL')
      >> acl.allow('system.Everyone', 'View')
      >> acl.deny('system.Everyone', 'Edit')

      >> list(acl)
      [(Allow, 'system.Everyone', ('View',)),
       (Deny, 'system.Everyone', ('Edit',))]

    """

    def __init__(self, name):
        self.name = name

    def get(self, typ, role):
        for r in self:
            if r[0] == typ and r[1] == role:
                return r

        return None

    def allow(self, role, *permissions):
        """ Give permissions to role """

        if not isinstance(role, basestring):
            role = role.id

        rec = self.get(Allow, role)
        if rec is None:
            rec = [Allow, role, set()]
            self.append(rec)

        if rec[2] is ALL_PERMISSIONS:
            return

        if ALL_PERMISSIONS in permissions:
            rec[2] = ALL_PERMISSIONS
        else:
            rec[2].update(permissions)

    def deny(self, role, *permissions):
        """ Deny permissions for role """

        if not isinstance(role, basestring):
            role = role.id

        rec = self.get(Deny, role)
        if rec is None:
            rec = [Deny, role, set()]
            self.append(rec)

        if rec[2] is ALL_PERMISSIONS:
            return

        if ALL_PERMISSIONS in permissions:
            rec[2] = ALL_PERMISSIONS
        else:
            rec[2].update(permissions)

    def unset(self, role, *permissions):
        """ Unset any previously defined permissions """
        for perm in permissions:
            for rec in self:
                if role is not None and rec[1] != role:
                    continue

                if rec[2] is ALL_PERMISSIONS or perm is ALL_PERMISSIONS:
                    rec[2] = set()
                else:
                    if perm in rec[2]:
                        rec[2].remove(perm)

        records = []
        for rec in self:
            if rec[2]:
                records.append(rec)
        self[:] = records


DEFAULT_ACL = ACL('default_por_acl') 
DEFAULT_ACL.allow('system.Everyone', 'view_anon')
DEFAULT_ACL.allow('role:administrator', ALL_PERMISSIONS)

#/home
DEFAULT_ACL.allow(Authenticated, 'view_home')
DEFAULT_ACL.allow('role:redturtle_developer', 'search')

#/manage_svn
DEFAULT_ACL.allow('role:redturtle_developer', 'manage_svn')

#/view_iterations
DEFAULT_ACL.allow('role:local_developer', 'view_iterations')
DEFAULT_ACL.allow('role:local_project_manager', 'view_iterations')
DEFAULT_ACL.allow('role:internal_developer', 'view_iterations')
DEFAULT_ACL.allow('role:secretary', 'view_iterations')
DEFAULT_ACL.allow('role:project_manager', 'view_iterations')

#/manage_iterations and /generate_iterations
DEFAULT_ACL.allow('role:local_project_manager', 'manage_iterations')
DEFAULT_ACL.allow('role:secretary', 'manage_iterations')
DEFAULT_ACL.allow('role:project_manager', 'manage_iterations')

#/add_entry
DEFAULT_ACL.allow('role:local_developer', 'add_entry')
DEFAULT_ACL.allow('role:local_project_manager', 'add_entry')
DEFAULT_ACL.allow('role:external_developer', 'add_entry')
DEFAULT_ACL.allow('role:internal_developer', 'add_entry')
DEFAULT_ACL.allow('role:secretary', 'add_entry')
DEFAULT_ACL.allow('role:project_manager', 'add_entry')

#/reports/index
DEFAULT_ACL.allow('role:local_project_manager', 'reports_index')
DEFAULT_ACL.allow('role:local_developer', 'reports_index')
DEFAULT_ACL.allow('role:secretary', 'reports_index')
DEFAULT_ACL.allow('role:project_manager', 'reports_index')

#/reports/report_custom
DEFAULT_ACL.allow('role:local_project_manager', 'reports_custom')
DEFAULT_ACL.allow('role:secretary', 'reports_custom')
DEFAULT_ACL.allow('role:project_manager', 'reports_custom')

#/reports/report_all_entries
DEFAULT_ACL.allow('role:local_project_manager', 'reports_all_entries')
DEFAULT_ACL.allow('role:local_developer', 'reports_all_entries')
DEFAULT_ACL.allow('role:secretary', 'reports_all_entries')
DEFAULT_ACL.allow('role:project_manager', 'reports_all_entries')

#/reports/report_state_change
DEFAULT_ACL.allow('role:secretary', 'reports_state_change')
DEFAULT_ACL.allow('role:project_manager', 'reports_state_change')
DEFAULT_ACL.allow('role:local_project_manager', 'reports_state_change')

#/reports/my_entries
DEFAULT_ACL.allow('role:local_developer', 'reports_my_entries')
DEFAULT_ACL.allow('role:local_project_manager', 'reports_my_entries')
DEFAULT_ACL.allow('role:internal_developer', 'reports_my_entries')
DEFAULT_ACL.allow('role:external_developer', 'reports_my_entries')
DEFAULT_ACL.allow('role:secretary', 'reports_my_entries')
DEFAULT_ACL.allow('role:project_manager', 'reports_my_entries')

#/backlog
DEFAULT_ACL.allow('role:local_developer', 'view_backlog')
DEFAULT_ACL.allow('role:local_project_manager', 'view_backlog')
DEFAULT_ACL.allow('role:secretary', 'view_backlog')

CRUD_ACL = ACL('por_crud_acl')
CRUD_ACL.allow('role:administrator', ALL_PERMISSIONS)
CRUD_ACL.allow('role:redturtle_developer', 'search')

#Customer
CRUD_ACL.allow(Authenticated, 'customer_listing')

ONLY_PROJECT_ROLES = set([u'internal_developer', u'external_developer', u'customer'])

gsm = zope.component.getGlobalSiteManager()
log = logging.getLogger('penelope')


@cache_region('calculate_matrix')
def __calculate_matrix__(user_id):
    """
    Return tuple of global and local roles in following format::
    >>> __calculate_matrix__(140)
    (set(['secretary'], {'fta': set(['internal_developer'])})
    """
    from penelope.core.models import User, DBSession, Project
    user = DBSession.query(User).get(user_id)
    global_roles = set(user.roles_names)
    local_roles = {}
    if 'administrator' in global_roles:
        log.debug("User: %s.\nGlobal roles: %s.\nLocal roles: %s" % (user, global_roles, local_roles))
        return global_roles, local_roles

    def add_local_role(project_id, role_name):
        if not project_id in local_roles:
            local_roles[project_id] = set()
        local_roles[project_id].add(role_name)

    for group in user.groups:
        for role in group.roles_names:
            add_local_role(group.project_id, role)
    for project in DBSession().query(Project.id).filter(Project.manager == user):
        add_local_role(project.id, u'project_manager')

    # extract global_roles from local_roles:
    roles_from_projects = set([item for sublist in local_roles.values() for item in sublist])

    # if user is an internal_developer in one of the projects add:
    if u'internal_developer' in roles_from_projects:
        global_roles.add(u'local_developer')

    # if user is a project_manager in one of the projects add:
    if u'project_manager' in roles_from_projects:
        global_roles.add(u'local_project_manager')

    log.debug("User: %s.\nGlobal roles: %s.\nLocal roles: %s" % (user, global_roles, local_roles))
    return global_roles, local_roles


def invalidate_user_calculate_matrix(target, value, initiator):
    log.debug("Cache invalidated for %s" % target)
    region_invalidate(__calculate_matrix__, 'calculate_matrix', target.id)

def invalidate_users_calculate_matrix(target, value, initiator):
    log.debug("Cache invalidated for %s" % value)
    region_invalidate(__calculate_matrix__, 'calculate_matrix', value.id)


class GenericRoles(object):
    def __init__(self, context, user):
        self.context = context
        self.user = user
        self.global_roles, self.local_roles = self.roles_matrix()
        self.roles = self.global_roles.copy()

    def roles_matrix(self):
        return __calculate_matrix__(self.user.id)

    def get_roles(self):
        return self.roles

gsm.registerAdapter(GenericRoles, (Interface, IUser), IRoleFinder)


class DublincoreRelatedRoles(GenericRoles):
    def __init__(self, context, user):
        super(DublincoreRelatedRoles, self).__init__(context, user)
        if getattr(self.context, 'author', None) == self.user:
            self.roles.add('owner')

gsm.registerAdapter(DublincoreRelatedRoles, (IDublinCore, IUser), IRoleFinder)

class ProjectRelatedRoles(DublincoreRelatedRoles):
    def add_project_roles(self, project):
        if project.manager == self.user:
            self.roles.add('project_manager')
        roles = self.local_roles.get(project.id, [])
        for role in roles:
            self.roles.add(role)

    def get_roles(self):
        self.roles -= ONLY_PROJECT_ROLES
        if self.context.project:
            self.add_project_roles(self.context.project)
        return self.roles

gsm.registerAdapter(ProjectRelatedRoles, (IProjectRelated, IUser), IRoleFinder)


class CustomerRelatedRoles(ProjectRelatedRoles):
    def get_roles(self):
        self.roles -= ONLY_PROJECT_ROLES
        for project in self.context.projects:
            self.add_project_roles(project)
        return self.roles

gsm.registerAdapter(CustomerRelatedRoles, (ICustomer, IUser), IRoleFinder)


class UserRoles(GenericRoles):
    def get_roles(self):
        if self.context == self.user:
            self.roles.add('owner')
        return self.roles

gsm.registerAdapter(UserRoles, (IUser, IUser), IRoleFinder)



# -*- coding: utf-8 -*-

import hashlib
import string
import random
import re
import urllib
import datetime

from json import loads
from copy import deepcopy
from zope.interface import implements
from zope.component import getMultiAdapter
from plone.i18n.normalizer import idnormalizer
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Authenticated
from pyramid.threadlocal import get_current_request
from repoze.workflow import get_workflow
from repoze.workflow import WorkflowError

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, Float
from sqlalchemy import String, Date
from sqlalchemy import Unicode
from sqlalchemy import Boolean
from sqlalchemy import Sequence
from sqlalchemy import Table
from sqlalchemy import event, func, and_
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm import deferred
from sqlalchemy.orm import column_property
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm.properties import ColumnProperty

from penelope.core.security.acl import CRUD_ACL, ACL, IRoleFinder
from penelope.core.security.acl import invalidate_user_calculate_matrix, invalidate_users_calculate_matrix
from penelope.core import html2text
from penelope.core.models import Base, DBSession, dublincore, workflow, classproperty
from penelope.core.models.interfaces import ICustomerRequest, IRoleable, IProjectRelated
from penelope.core.models.interfaces import IProject, IUser, IPorModel, ICustomer, IRole
from penelope.core.models.interfaces import IApplication, ITrac, ISVN, IGoogleDocs
from penelope.core.models.interfaces import ITracReport, IGenericApp, IContract, IKanbanBoard


role_assignments = Table('role_assignments', Base.metadata,
    Column('principal_id', Integer, ForeignKey('principals.id')),
    Column('role_id', String, ForeignKey('roles.id'))
)

group_assignments = Table('group_assignments', Base.metadata,
    Column('group_id', Integer, ForeignKey('groups.id')),
    Column('user_id', Integer, ForeignKey('users.id'))
)

favorite_projects = Table('favorite_projects', Base.metadata,
    Column('project_id', String, ForeignKey('projects.id')),
    Column('user_id', Integer, ForeignKey('users.id'))
)

kanban_projects = Table('kanban_projects', Base.metadata,
    Column('project_id', String, ForeignKey('projects.id')),
    Column('kanban_id', Integer, ForeignKey('kanban_boards.id'))
)


class GlobalConfig(Base):
    implements(IPorModel)

    __tablename__ = "global_config"
    __acl__ = ACL('por_global_config_acl')
    __acl__.deny('system.Everyone', ALL_PERMISSIONS)

    id = Column(Integer, primary_key=True)
    active_iteration_url = Column(Unicode)

    def cost_per_day(self, day):
        for cost in self.costs:
            if day >= cost.date:
                return cost


class Principal(Base):
    implements(IPorModel)

    __tablename__ = "principals"
    __acl__ = deepcopy(CRUD_ACL)

    id = Column(Integer, Sequence("principals_id"), primary_key=True)


class Role(Base):
    implements(IRole)

    __tablename__ = "roles"
    __acl__ = deepcopy(CRUD_ACL)

    id = Column(String, primary_key=True)
    name = Column(String(64), unique=True)

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return "<Role id=%s name=%s>" % (self.id, self.name)


def new_role_created(mapper, connection, target):
    target.id = idnormalizer.normalize(target.name)

event.listen(Role, "before_insert", new_role_created)


class LowercaseComparator(ColumnProperty.Comparator):
    def __eq__(self, other):
        return func.lower(self.__clause_element__()) == func.lower(other)


class User(Principal):
    implements(IRoleable, IUser)

    __tablename__ = 'users'

    @classproperty
    def __acl__(cls):
        acl = deepcopy(CRUD_ACL)
        if bool(cls.active):
            acl.allow('role:owner', 'edit')
            acl.allow('role:owner', 'view')
        return acl

    id = Column(Integer, ForeignKey(Principal.id),
        Sequence("principals_id"),
        primary_key=True)

    fullname = Column(Unicode)
    __mapper_args__ = {'order_by': func.substring(fullname, '([^[:space:]]+)(?:,|$)')}
    email = column_property(Column('email', Unicode, unique=True),
                            comparator_factory=LowercaseComparator)
    svn_login = deferred(Column(Unicode, unique=True, nullable=False))
    phone = deferred(Column(String(20)))
    mobile = deferred(Column(String(20)))
    password = deferred(Column(String))
    salt = deferred(Column(String(12)))
    roles = relationship(Role, secondary=role_assignments, backref="users")
    gdata_auth_token = deferred(Column(String(64)))
    active = deferred(Column(Boolean, nullable=False, default=True))

    @hybrid_property
    def login(self):
        return self.email

    def generate_salt(self):
        return ''.join(random.sample(string.letters, 12))

    def encrypt(self, password):
        #force everything to become string:
        password = password.encode('utf8','ignore')
        return str(hashlib.sha1(password + str(self.salt)).hexdigest())

    def set_password(self, password):
        self.salt = self.generate_salt()
        self.password = self.encrypt(password)

    def check_password(self, password):
        if not self.active:
            return False
        if not self.password:
            return False
        return self.encrypt(password) == self.password

    def add_openid(self, openid):
        openid = OpenId(openid=openid)
        self.openids.append(openid)

    def cost_per_day(self, day):
        for cost in self.costs:
            if day >= cost.date:
                return cost

    @property
    def email_domains(self):
        return re.compile("@([\w.]+)").findall(self.email)

    @property
    def gdata_token_status(self):
        return self.gdata_auth_token and 'Active' or 'Not active'

    @hybrid_property
    def roles_names(self):
        return [a.id.lower() for a in self.roles]

    def roles_in_context(self, context=None):
        return getMultiAdapter((context, self), IRoleFinder).get_roles()

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.fullname or self.email

    def __repr__(self):
        return "<User login=%s>" % self.login


def user_email(mapper, connection, target):
    if target.email:
        target.email = target.email.lower().strip()
        if not target.svn_login:
            target.svn_login = target.email

event.listen(User, "before_insert", user_email)
event.listen(User, "before_update", user_email)
event.listen(User.roles, "append", invalidate_user_calculate_matrix)
event.listen(User.roles, "remove", invalidate_user_calculate_matrix)


class PasswordResetToken(Base):
    __tablename__ = 'password_reset_tokens'
    user_id = Column(Integer, ForeignKey('users.id'), primary_key=True)
    user = relationship(User, uselist=False)
    token = Column(String, unique=True)

    @property
    def id(self):
        return self.token


class OpenId(Base):
    implements(IPorModel)

    __tablename__ = 'openids'
    __acl__ = deepcopy(CRUD_ACL)
    id = Column(Integer, primary_key=True)
    openid = Column(String)
    user_id = Column(Integer, ForeignKey('users.id'))
    user = relationship(User, uselist=False, backref=backref('openids', order_by=id))

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.openid

    def __repr__(self):
        return "<OpenID openid=%s>" % self.openid


class Cost(Base):
    implements(IPorModel)

    __tablename__ = 'costs'
    __acl__ = deepcopy(CRUD_ACL)
    id = Column(Integer, primary_key=True)
    date = Column(Date, index=True)
    amount = Column(Float(precision=2))
    user_id = Column(Integer, ForeignKey('users.id'))
    user = relationship(User, uselist=False, backref=backref('costs', order_by=date.desc()))
    global_config_id = Column(Integer, ForeignKey('global_config.id'))
    global_config = relationship(GlobalConfig, uselist=False, backref=backref('costs', order_by=date.desc()))

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        name = self.user and self.user.fullname or 'Company'
        return u"%s's cost for %s" % (name, self.date)

    def __repr__(self):
        return "<Cost user=%s>" % self.user_id


class Customer(dublincore.DublinCore, Base):
    implements(ICustomer)

    __tablename__ = 'customers'
    __acl__ = deepcopy(CRUD_ACL)
    #view
    __acl__.allow('role:customer', 'view')
    __acl__.allow('role:external_developer', 'view')
    __acl__.allow('role:internal_developer', 'view')
    __acl__.allow('role:secretary', 'view')
    __acl__.allow('role:project_manager', 'view')
    #metadata
    __acl__.allow('role:customer', 'metadata')
    __acl__.allow('role:project_manager', 'metadata')
    #timeentries
    __acl__.allow('role:secretary', 'time_entries')
    __acl__.allow('role:project_manager', 'time_entries')
    #edit
    __acl__.allow('role:project_manager', 'edit')
    #new
    __acl__.allow('role:project_manager', 'new')
    #delete
    __acl__.allow('role:project_manager', 'delete')

    id = Column(String, primary_key=True)
    name = Column(Unicode, unique=True)
    __mapper_args__ = {'order_by': name}

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.name

    def add_project(self, project):
        if [proj for proj in self.projects if proj.name == project.name]:
            raise AttributeError('Duplicated project name. %s already exists' % project.name)
        self.projects.append(project)

    @property
    def color(self):
        return hashlib.md5(self.name.encode('utf8','ignore')).hexdigest()[:6]


def new_customer_created(mapper, connection, target):
    target.id = idnormalizer.normalize(target.name)

event.listen(Customer, "before_insert", new_customer_created)
event.listen(Customer, "before_insert", dublincore.dublincore_insert)
event.listen(Customer, "before_update", dublincore.dublincore_update)


class Project(dublincore.DublinCore, Base):

    implements(IProject, IProjectRelated)

    project_related_label = 'Project'
    project_related_id = ''

    __acl__ = deepcopy(CRUD_ACL)
    #listing
    __acl__.allow(Authenticated, 'listing')
    #view
    __acl__.allow('role:customer', 'view')
    __acl__.allow('role:external_developer', 'view')
    __acl__.allow('role:internal_developer', 'view')
    __acl__.allow('role:secretary', 'view')
    __acl__.allow('role:project_manager', 'view')
    #metadata
    __acl__.allow('role:secretary', 'metadata')
    __acl__.allow('role:project_manager', 'metadata')
    #add
    __acl__.allow('role:project_manager', 'new')
    #edit
    __acl__.allow('role:project_manager', 'edit')
    #edit
    __acl__.allow('role:project_manager', 'delete')
    #estimations
    __acl__.allow('role:secretary', 'estimations')
    __acl__.allow('role:project_manager', 'estimations')
    __acl__.allow('role:internal_developer', 'estimations')
    #view all entries
    __acl__.allow('role:secretary', 'reports_all_entries_for_project')
    __acl__.allow('role:project_manager', 'reports_all_entries_for_project')
    __acl__.allow('role:internal_developer', 'reports_all_entries_for_project')
    #contract
    __acl__.allow('role:project_manager', 'view_contracts')
    __acl__.allow('role:secretary', 'view_contracts')
    __acl__.allow('role:internal_developer', 'view_contracts')
    __acl__.allow('role:project_manager', 'add_contracts')
    __acl__.allow('role:secretary', 'add_contracts')
    #groups
    __acl__.allow('role:project_manager', 'view_groups')
    __acl__.allow('role:project_manager', 'add_groups')
    #customer_requests
    __acl__.allow('role:project_manager', 'list_customer_request')
    __acl__.allow('role:project_manager', 'add_customer_request')
    __acl__.allow('role:internal_developer', 'list_customer_request')
    __acl__.allow('role:internal_developer', 'add_customer_request')
    __acl__.allow('role:secretary', 'list_customer_request')
    __acl__.allow('role:secretary', 'add_customer_request')
    __acl__.allow('role:customer', 'list_customer_request')
    __acl__.allow('role:external_developer', 'list_customer_request')

    __tablename__ = 'projects'

    id = Column(String, primary_key=True)
    name = Column(Unicode)
    __mapper_args__ = {'order_by': name}
    activated = Column(Boolean, nullable=False, default=True)

    # quality end date fields
    completion_date = Column(Date)
    assistance_date = Column(Date)
    test_date = Column(Date)
    inception_date = Column(Date)

    customer_id = Column(String, ForeignKey('customers.id'))
    customer = relationship(Customer, uselist=False, backref=backref('projects'))
    manager_id = Column(Integer, ForeignKey('users.id'))
    manager = relationship(User,
                           primaryjoin=and_(manager_id == User.id,
                                            User.active == True),
                           uselist=False,
                           backref=backref('project_manager'),
                           order_by=func.substring(User.fullname,'([^[:space:]]+)(?:,|$)'))
    favorite_users = relationship(User, secondary=favorite_projects, backref="favorite_projects")

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        if self.customer and self.name:
            return '%s - %s' % (unicode(self.customer), self.name)
        else:
            return self.name or ''

    @property
    def project(self):
        "compatibility with IProjectRelated"
        return self

    def dashboard_apps(self):
        """ return only svn and trac """
        return [a for a in self.applications if a.application_type in [TRAC, SVN]]

    def add_application(self, application):
        if [app for app in self.applications if app.name == application.name]:
            raise AttributeError('Duplicated application name. %s already exists' % application.name)
        self.applications.append(application)

    def add_customer_request(self, customer_request):
        if [cr for cr in self.customer_requests if cr.name == customer_request.name]:
            raise AttributeError('Duplicated customer request name. %s already exists' % customer_request.name)
        self.customer_requests.append(customer_request)

    def add_group(self, group):
        self.groups.append(group)

    @property
    def tracs(self):
        for app in self.applications:
            if ITrac.providedBy(app):
                yield app

    def get_number_of_tickets_per_cr(self):
        from penelope.core.models.tickets import ticket_store
        return ticket_store.get_number_of_tickets_per_cr(self)

    @hybrid_method
    def users_favorite(self, user):
        return (favorite_projects.c.user_id == user.id) & (favorite_projects.c.project_id == self.id)

    @hybrid_property
    def active(self):
        return (self.activated != False) & (self.activated != None)

    def contracts_by_state(self):
        return DBSession().query(Contract.id, Contract.name, Contract.workflow_state)\
                          .join(Project).filter(Project.id==self.id)\
                          .order_by(Contract.active.desc())\
                          .order_by(Contract.modification_date.desc())\
                          .order_by(Contract.name)


def new_project_created(mapper, connection, target):
    project_id_candidate = target.id or target.name
    target.id = idnormalizer.normalize(project_id_candidate)
    target.inception_date = datetime.datetime.now()

event.listen(Project, "before_insert", new_project_created)
event.listen(Project, "before_insert", dublincore.dublincore_insert)
event.listen(Project, "before_update", dublincore.dublincore_update)
event.listen(Project.manager, "append", invalidate_users_calculate_matrix)
event.listen(Project.manager, "remove", invalidate_users_calculate_matrix)


GOOGLE_DOCS = 'google docs'
SVN = 'svn'
TRAC = 'trac'
TRAC_REPORT = 'trac report'
GENERIC_APP = 'generic'


class Application(dublincore.DublinCore, Base):
    implements(IProjectRelated, IApplication)

    project_related_label = 'Applications'
    project_related_id = 'applications'

    __tablename__ = 'applications'


    @classproperty
    def __acl__(self):
        acl = deepcopy(CRUD_ACL)

        # listing
        acl.allow('role:project_manager', 'listing')
        acl.allow('role:customer', 'listing')
        acl.allow('role:internal_developer', 'listing')
        acl.allow('role:external_developer', 'listing')
        acl.allow('role:secretary', 'listing')

        # project manager
        acl.allow('role:project_manager', 'view')
        acl.allow('role:project_manager', 'edit')
        acl.allow('role:project_manager', 'delete')
        acl.allow('role:project_manager', 'new')

        # dynamically set by CRUD form
        for _acl in self.acl:
            acl.allow('role:%s' % _acl.role_id, _acl.permission_name)

        return acl


    id = Column(Integer, primary_key=True)
    position = Column(Integer, nullable=False, default=-1)
    name = Column(Unicode)
    trac_name = deferred(Column(String(128)))
    svn_name = deferred(Column(String(128)))
    description = deferred(Column(Unicode))
    project_id = Column(String, ForeignKey('projects.id'))

    project = relationship(Project, uselist=False,
                           backref=backref('applications',
                                           collection_class=ordering_list('position'),
                                           order_by=position))

    api_uri = deferred(Column(String, default=''))

    application_type = Column(String(50))
    __mapper_args__ = {'polymorphic_on': application_type,
                       'polymorphic_identity': ''}

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.name

    def application_uri(self, request=None):
        if self.api_uri.startswith('trac://') or self.api_uri.startswith('svn://'):
            new_api_uri = update_trac_svn_uri(None, self.api_uri, None, None)
            self.api_uri = new_api_uri
        return self.api_uri

    def get_icon(self):
        return {
                'google docs': 'icon-file-text-alt',
                'trac': 'icon-bug',
                'svn': 'icon-code-fork',
                }.get(self.application_type, 'icon-folder-close-alt')

    def get_color(self):
        return {
                'google docs': 'btn-primary',
                'trac': 'btn-danger',
                'svn': 'btn-warning',
                }.get(self.application_type, 'btn-inverse')


class ApplicationACL(Base):
    __tablename__ = 'applications_acl'

    application_id = Column(Integer, ForeignKey('applications.id', ondelete='cascade'), primary_key=True)
    role_id = Column(String, ForeignKey('roles.id', ondelete='cascade'), primary_key=True)
    permission_name = Column(String, primary_key=True)

    project = relationship(Application, uselist=False, backref=backref('acl',
                                                                       cascade="all, delete-orphan",
                                                                       passive_deletes=True))
    role = relationship(Role, uselist=False)

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return '(%s, %s, %s)' % (self.application_id, self.role_id, self.permission_name)



def modify_application_type(mapper, connection, target):
    request = get_current_request()
    if request:
        _id = target.id or ''
        application_type = str(request.params.get(u'%s-%s-application_type' % (
                               target.__class__.__name__, _id, ),''))
        if application_type:
            target.application_type = application_type

def update_trac_svn_uri(target, value, oldvalue, initiator):
    request = get_current_request()
    if request:
        proto = request.environ.get('wsgi.url_scheme')
        http_host = request.environ.get('HTTP_HOST')
        svnurl = request.registry.settings.get('por.svn.url')
        if svnurl.endswith('/'):
            svnurl = svnurl[:-1]
        if not oldvalue and value:
            if value.startswith('svn://'):
                svn_id = value[6:]
                return '%s/%s' % (svnurl, svn_id)
            elif value.startswith('trac://'):
                trac_id = value[7:]
                return '%s://%s/trac/%s' % (proto, http_host, trac_id)
    return value

def update_app_position(mapper, connection, target):
    for n, app in enumerate(target.project.applications):
        if app.id == target.id:
            if target.position == -1: #  keep last 
                app = target.project.applications.pop(n)
                DBSession().query(Project).get(app.project_id).applications.append(app)
            elif n != target.position:
                app = target.project.applications.pop(n)
                DBSession().query(Project).get(app.project_id).applications.insert(target.position, app)

            target.project.applications.reorder()
            break

def default_app_position(mapper, connection, target):
    if not target.position:
        app_len = len(DBSession().query(Project).get(target.project_id).applications)
        target.position = app_len + 100


def create_initial_application_acl(mapper, connection, target):
    if target.application_type == SVN:
        acl_rules = [
                    ('internal_developer', 'edit'),
                    ('internal_developer', 'view'),
                    ('external_developer', 'edit'),
                    ('external_developer', 'view'),
                    ]
    else:
        acl_rules = [
                    ('internal_developer', 'view'),
                    ('external_developer', 'view'),
                    ('secretary', 'view'),
                    ('secretary', 'edit'),
                    ]

    if target.application_type == 'trac':
        acl_rules.append(('customer', 'view'))

    for role_id, permission_name in acl_rules:
        acl = DBSession.query(ApplicationACL).get((target.id, role_id, permission_name))
        if not acl:
            acl = ApplicationACL(application_id=target.id,
                                 role_id=role_id,
                                 permission_name=permission_name)
            DBSession.add(acl)
        else:
            # XXX this should not happen.
            pass



event.listen(Application.api_uri, "set", update_trac_svn_uri, retval=True)
event.listen(Application, "before_insert", modify_application_type, propagate=True)
event.listen(Application, "before_insert", default_app_position, propagate=True)
event.listen(Application, "before_insert", dublincore.dublincore_insert, propagate=True)
event.listen(Application, "after_insert", create_initial_application_acl, propagate=True)
event.listen(Application, "before_update", modify_application_type, propagate=True)
event.listen(Application, "before_update", dublincore.dublincore_update, propagate=True)
event.listen(Application, "before_update", update_app_position, propagate=True)


class GenericApp(Application):

    implements(IGenericApp)
    __mapper_args__ = {'polymorphic_identity': GENERIC_APP}


class Trac(Application):

    implements(ITrac)
    __mapper_args__ = {'polymorphic_identity': TRAC}


class TracReport(Application):

    implements(ITracReport)
    __mapper_args__ = {'polymorphic_identity': TRAC_REPORT}


class Subversion(Application):

    implements(ISVN)
    __mapper_args__ = {'polymorphic_identity': SVN}


class GoogleDoc(Application):

    implements(IGoogleDocs)
    __mapper_args__ = {'polymorphic_identity': GOOGLE_DOCS}


class Contract(dublincore.DublinCore, workflow.Workflow, Base):
    implements(IContract, IProjectRelated)

    project_related_label = 'Contracts'
    project_related_id = 'contracts'

    __tablename__ = 'contracts'
    __acl__ = deepcopy(CRUD_ACL)
    #view
    __acl__.allow('role:project_manager', 'view')
    __acl__.allow('role:secretary', 'view')
    __acl__.allow('role:internal_developer', 'view')
    #workflow
    __acl__.allow('role:secretary', 'workflow')
    __acl__.allow('role:project_manager', 'workflow')
    __acl__.allow('role:project_manager', 'workflow_activate')
    __acl__.allow('role:project_manager', 'workflow_deactivate')
    __acl__.allow('role:project_manager', 'workflow_achieve')
    __acl__.allow('role:secretary', 'workflow_achieve')
    __acl__.allow('role:project_manager', 'workflow_unachieve')
    __acl__.allow('role:secretary', 'workflow_unachieve')
    #add
    __acl__.allow('role:project_manager', 'new')
    __acl__.allow('role:secretary', 'new')
    #edit
    __acl__.allow('role:project_manager', 'edit')
    __acl__.allow('role:secretary', 'edit')

    id = Column(String, primary_key=True)
    name = Column(Unicode, nullable=False)
    description = deferred(Column(Unicode))
    days = Column(Float(precision=2), nullable=False, default=0)
    amount = Column(Float(precision=2))
    contract_number = Column(Unicode)
    start_date = Column(Date)
    end_date = Column(Date)

    project_id = Column(String, ForeignKey('projects.id'), nullable=False)
    project = relationship(Project, uselist=False, backref=backref('contracts', order_by=id))

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.name

    @hybrid_property
    def active(self):
        return (self.workflow_state == 'active')


def new_contract_created(mapper, connection, target):
    contract_id_candidate = '%s_%s' % (target.project_id, target.name)
    target.id = idnormalizer.normalize(contract_id_candidate)

event.listen(Contract, "before_insert", new_contract_created)
event.listen(Contract, "before_insert", dublincore.dublincore_insert)
event.listen(Contract, "before_update", dublincore.dublincore_update)


class CustomerRequest(dublincore.DublinCore, workflow.Workflow, Base):
    implements(ICustomerRequest, IProjectRelated)

    project_related_label = 'Customer requests'
    project_related_id = 'customer_requests'

    __tablename__ = 'customer_requests'
    __acl__ = deepcopy(CRUD_ACL)
    #view
    __acl__.allow('role:project_manager', 'view')
    __acl__.allow('role:secretary', 'view')
    __acl__.allow('role:internal_developer', 'view')
    __acl__.allow('role:customer', 'view')
    #estimations
    __acl__.allow('role:secretary', 'estimations')
    __acl__.allow('role:project_manager', 'estimations')
    __acl__.allow('role:internal_developer', 'estimations')
    #workflow
    __acl__.allow('role:project_manager', 'workflow')
    #add
    __acl__.allow('role:project_manager', 'new')
    __acl__.allow('role:internal_developer', 'new')
    #edit
    __acl__.allow('role:project_manager', 'edit')
    __acl__.allow('role:internal_developer', 'edit')
    #contract view/edit
    __acl__.allow('role:secretary', 'contract')
    __acl__.allow('role:project_manager', 'contract')
    #delete
    __acl__.allow('role:project_manager', 'delete')

    id = Column(String, primary_key=True)
    uid = Column(Integer)
    name = Column(Unicode)
    description = deferred(Column(Unicode))
    project_id = Column(String, ForeignKey('projects.id'))
    project = relationship(Project, uselist=False, backref=backref('customer_requests'))
    contract_id = Column(String, ForeignKey('contracts.id'))
    contract = relationship(Contract, uselist=False, backref=backref('customer_requests'))
    old_contract_name = Column(Unicode)
    filler = Column(Boolean, nullable=False, default=False)

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        if self.contract:
            return '%s [%s]' % (self.name, self.contract)
        else:
            return self.name

    @hybrid_property
    def active(self):
        return (self.workflow_state == 'estimated' or self.workflow_state == 'created')

    def get_tickets(self, request=None):
        from penelope.core.models.tickets import ticket_store
        return ticket_store.get_tickets_for_request(customer_request=self)

    def add_ticket_url(self, request):
        for trac in self.project.tracs:
            trac_url = trac.application_uri(request)
            if self.description:
                desc = html2text.html2text(self.description).encode('utf8')
            else:
                desc = ''
            params = urllib.urlencode({'summary':str(self), 
                                       'type': 'task',
                                       'customerrequest': self.id,
                                       'description': desc})
            return '%s/newticket?%s' % (trac_url, params)

    @property
    def estimation_days(self):
        if self.filler and self.contract:
            other_crs = DBSession().query(CustomerRequest)\
                                   .filter_by(filler=False)\
                                   .filter_by(contract_id=self.contract_id)
            filler = self.contract.days - sum([cr.estimation_days for cr in other_crs])
            if filler < 0:
                return 0
            else:
                return filler
        return sum([a.days for a in self.estimations], float())

    @property
    def timeentries_days(self):
        from penelope.core.lib.helpers import timedelta_as_work_days
        hours = sum([a.hours for a in self.time_entries], datetime.timedelta())
        return timedelta_as_work_days(hours)


def after_customer_request_flushed(session, flush_context, instances):
    """ generate unique id per project_id """
    def last_cr4project(obj):
        project_id = idnormalizer.normalize(obj.project_id or obj.project.name)
        last = session.query(CustomerRequest.uid)\
                    .filter_by(project_id=project_id)\
                    .order_by(CustomerRequest.uid.desc()).first()
        if last: new_uid = last.uid + 1
        else: new_uid = 1
        return project_id, new_uid

    last = {}
    for obj in session.new:
        if isinstance(obj, CustomerRequest):
            project_id, new_uid = last_cr4project(obj)
            if not project_id in last.keys():
                last[project_id] = new_uid 
            else:
                last[project_id] = last[project_id] + 1
            obj.uid = last[project_id]
            obj.id = "%s_%s" % (project_id, last[project_id])


event.listen(DBSession, "before_flush", after_customer_request_flushed)
event.listen(CustomerRequest, "before_insert", dublincore.dublincore_insert)
event.listen(CustomerRequest, "before_update", dublincore.dublincore_update)



class Estimation(Base):
    __tablename__ = "estimations"
    __acl__ = deepcopy(CRUD_ACL)
    #view
    __acl__.allow('role:project_manager', 'view')
    __acl__.allow('role:secretary', 'view')
    __acl__.allow('role:internal_developer', 'view')
    #add
    __acl__.allow('role:project_manager', 'new')
    __acl__.allow('role:internal_developer', 'new')
    #edit
    __acl__.allow('role:project_manager', 'edit')
    __acl__.allow('role:internal_developer', 'edit')

    id = Column(Integer, primary_key=True)
    days = Column(Float(precision=2), nullable=False)
    person_type = Column(Unicode, nullable=False)
    customer_request_id = Column(String, ForeignKey('customer_requests.id'))
    customer_request = relationship(CustomerRequest, backref=backref('estimations'))


def check_cr_for_estimation(mapper, connection, target):
    state = ''
    cr = DBSession().query(CustomerRequest).get(target.customer_request_id)

    if target in cr.estimations and len(cr.estimations) == 1 and cr.workflow_state != 'created': # removing last estimation
        state = 'created'
    elif len(cr.estimations)==0 and cr.workflow_state != 'estimated':
        state = 'estimated'

    request = get_current_request()
    if state and request:
        workflow = get_workflow(cr, 'CustomerRequest')
        try:
            workflow.transition_to_state(cr, request, state, skip_same=True)
        except WorkflowError:
            pass


event.listen(Estimation, "before_insert", check_cr_for_estimation)
event.listen(Estimation, "before_delete", check_cr_for_estimation)


class Group(Principal):
    implements(IRoleable, IProjectRelated)

    project_related_label = 'Groups'
    project_related_id = 'configuration'

    __tablename__ = "groups"
    __acl__ = deepcopy(CRUD_ACL)
    #view
    __acl__.allow('role:project_manager', 'view')
    #add
    __acl__.allow('role:project_manager', 'new')
    #edit
    __acl__.allow('role:project_manager', 'edit')
    __acl__.allow('role:project_manager', 'manage_roles')
    #delete
    __acl__.allow('role:project_manager', 'delete')

    id = Column(Integer, ForeignKey(Principal.id),
        Sequence("principals_id"),
        primary_key=True)
    project_id = Column(String, ForeignKey('projects.id'))
    project = relationship(Project, uselist=False, backref=backref('groups'))
    roles = relationship(Role, secondary=role_assignments, backref="groups")
    users = relationship(User,
                         secondary=group_assignments,
                         backref="groups",
                         order_by=func.substring(User.fullname,'([^[:space:]]+)(?:,|$)'))

    def add_user(self, user):
        if [us for us in self.users if us.login == user.login]:
            raise AttributeError('Dupplicated username. %s already exists' % user.login)
        self.users.append(user)

    @hybrid_property
    def roles_names(self):
        if self.roles:
            return [a.id.lower() for a in self.roles]
        else:
            return ['[No role]']

    @roles_names.expression
    def roles_names_expr(cls):
        return Role.name

    def __repr__(self):
        return "<Group id=%d>" % self.id

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return "[%s] %s" % (self.project.name, ','.join(self.roles_names))

def wrap_users_before_group_delete(mapper, connection, target):
    for user in target.users:
        invalidate_user_calculate_matrix(user, None, None)

event.listen(Group.users, "append", invalidate_users_calculate_matrix)
event.listen(Group.users, "remove", invalidate_users_calculate_matrix)
event.listen(Group, "before_delete", wrap_users_before_group_delete)


class SavedQuery(dublincore.DublinCore, Base):
    implements(IPorModel)

    __tablename__ = 'saved_queries'

    id = Column(Integer, primary_key=True)
    query_name = Column(Unicode)
    report_name = Column(Unicode)
    query_string = Column(Unicode)

    def __repr__(self):
        return "<SavedQuery id=%d name=%s>" % (self.id, self.query_name)

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.query_name

    def full_url(self, request):
        return request.route_url('reports', traverse=[self.report_name]) + self.query_string

event.listen(SavedQuery, "before_insert", dublincore.dublincore_insert)
event.listen(SavedQuery, "before_update", dublincore.dublincore_update)


BACKLOG_PRIORITY_ORDER = 'priority'
BACKLOG_MODIFICATION_ORDER = 'modification'


class KanbanBoard(dublincore.DublinCore, Base):
    implements(IKanbanBoard)

    __tablename__ = 'kanban_boards'

    @classproperty
    def __acl__(self):
        acl = deepcopy(CRUD_ACL)

        acl.allow('role:redturtle_developer', 'new')
        acl.allow('role:redturtle_developer', 'listing')

        acl.allow('role:owner', 'view')
        acl.allow('role:owner', 'edit')
        acl.allow('role:owner', 'delete')

        # dynamically set by deform
        try:
            for _acl in self.acl:
                acl.allow(_acl.principal, _acl.permission_name)
        except TypeError: # if self is a class the acl attribute is not iterable
            pass
        return acl

    id = Column(Integer, primary_key=True)
    name = Column(Unicode, nullable=False)
    json = Column(Unicode)
    backlog_query = Column(Unicode, nullable=False)
    backlog_limit = Column(Integer, default=10, nullable=False)
    backlog_order = Column(Unicode, nullable=False)
    author = relationship(User,
                          uselist=False,
                          primaryjoin='KanbanBoard.author_id==User.id',
                          backref='kanban_boards')
    projects = relationship(Project, secondary=kanban_projects, backref="kanban_boards")

    def __repr__(self):
        return "<KanbanBoard id=%d>" % self.id

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.name

    def get_board_columns(self):
        columns = self.json and loads(self.json) or []
        return [str(len(c['tasks'])) for c in columns]

def create_initial_kanban_acl(mapper, connection, target):
    acl_rules = [
                 ('role:redturtle_developer', 'view'),
                ]

    for principal_id, permission_name in acl_rules:
        acl = KanbanACL(principal=principal_id,
                board_id=target.id,
                permission_name=permission_name)
        DBSession().add(acl)

event.listen(KanbanBoard, "after_insert", create_initial_kanban_acl, propagate=True)
event.listen(KanbanBoard, "before_insert", dublincore.dublincore_insert)
event.listen(KanbanBoard, "before_update", dublincore.dublincore_update)

class KanbanACL(Base):
    __tablename__ = 'kanban_acl'

    id = Column(Integer, primary_key=True)
    principal = Column(String)
    permission_name = Column(String)
    board_id = Column(Integer, ForeignKey(KanbanBoard.id))
    board = relationship(KanbanBoard, uselist=False, backref=backref('acl',
        cascade="all, delete-orphan",
        passive_deletes=True))

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return '(%s, %s, %s)' % (self.principal, self.board_id, self.permission_name)



# -*- coding: utf-8 -*-

from copy import deepcopy
import json

from pyramid.httpexceptions import HTTPFound
from pyramid import httpexceptions as exc
from pyramid_skins import SkinObject
from repoze.workflow import get_workflow, WorkflowError
from pyramid_formalchemy import events
from zope.interface import alsoProvides
import zope.component.event


from fa.bootstrap import actions

from penelope.core.backlog import Backlog
from penelope.core.models import dashboard
from penelope.core.models import DBSession
from penelope.core.forms import ModelView


def configurate(config):
    #custom view for project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='',
        attr='documentation',
        renderer='fa.bootstrap:templates/admin/show.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        renderer='penelope.core.forms:templates/project_listing.pt',
        attr='datatable',
        context='pyramid_formalchemy.resources.ModelListing',
        request_method='GET',
        permission='listing',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding an application to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_group',
        attr='add_group',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_group',
        attr='add_group',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding an application to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_application',
        attr='add_application',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_application',
        attr='add_application',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for configuration section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='edit',
        name='configuration',
        attr='configuration',
        renderer='penelope.core.forms:templates/configuration.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for documentation section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='metadata',
        name='metadata',
        attr='show',
        renderer='penelope.core.forms:templates/project.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for customer_requests section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='list_customer_request',
        name='customer_requests',
        attr='customer_requests',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for customer_requests section
    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='edit',
        name='customer_requests',
        attr='update_cr_states',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for contracts section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='contracts',
        attr='contracts',
        renderer='penelope.core.forms:templates/contracts.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)


    #custom view for tickets section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='tickets',
        attr='tickets',
        renderer='penelope.core.forms:templates/tickets.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for time_entries section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='time_entries',
        name='time_entries',
        attr='time_entries',
        renderer='penelope.core.forms:templates/time_entries.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for documentation section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='applications',
        attr='applications',
        renderer='penelope.core.forms:templates/configurate_apps.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding a customer request to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='add_customer_request',
        name='add_customer_request',
        attr='add_customer_request',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='add_customer_request',
        name='add_customer_request',
        attr='add_customer_request',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding a contract to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_contract',
        attr='add_contract',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_contract',
        attr='add_contract',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)


    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='toggle_favorite',
        attr='toggle_favorite',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='list_customer_request',
        name='list_customer_requests.json',
        renderer='json',
        attr='list_customer_requests',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

add_customer_request_actions = deepcopy(actions.defaults_actions['new_buttons'])
add_customer_request_actions[0].permission = 'add_customer_request'
add_customer_request_actions[1].permission = 'add_customer_request'
add_customer_request_actions[2].permission = 'add_customer_request'


class ProjectModelView(ModelView):
    actions_categories = ('buttons', 'tabs', 'subtabs')
    defaults_actions = deepcopy(actions.defaults_actions)

    def __init__(self, *args, **kwargs):
        super(ProjectModelView, self).__init__(*args, **kwargs)
        self.defaults_actions.update(show_buttons=actions.Actions(actions.edit))
        self.defaults_actions.update(listing_buttons=actions.Actions())
        self.defaults_actions.update(documentation_buttons=actions.Actions())
        self.defaults_actions['add_customer_request_buttons'] = add_customer_request_actions

    def delete(self):
        project = self.context.get_instance()
        request = self.request
        STOP = False
        if len(project.groups):
            request.add_message(u'Project has existing groups. Please remove them first.', type='danger')
            STOP = True
        if len(project.applications):
            request.add_message(u'Project has existing applications. Please remove them first.', type='danger')
            STOP = True
        if len(project.customer_requests):
            request.add_message(u'Project has existing customer_requests. Please remove them first.', type='danger')
            STOP = True
        if len(project.time_entries):
            request.add_message(u'Project has existing time entries. Please remove them first.', type='danger')
            STOP = True

        if STOP:
            raise exc.HTTPFound(location=request.fa_url('Project', project.id))
        else:
            request.fa_parent_url = lambda : request.fa_url(request.model_name)
            return self.force_delete()

    def add_group(self, *args, **kwargs):
        self.request.model_class = dashboard.Group
        self.request.model_name = dashboard.Group.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_group']
        return self.new()

    def add_application(self, *args, **kwargs):
        self.request.model_class = dashboard.Application
        self.request.model_name = dashboard.Application.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_application']
        return self.new()

    @actions.action('add_customer_request')
    def add_customer_request(self, *args, **kwargs):
        self.request.model_class = dashboard.CustomerRequest
        self.request.model_name = dashboard.CustomerRequest.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_customer_request']
        if self.request.method == 'POST':
            return self.create()

        fs = self.get_fieldset(suffix='Add')
        fs = fs.bind(session=self.session, request=self.request)

        event = events.BeforeRenderEvent(fs.model, self.request, fs=fs)
        alsoProvides(event, events.IBeforeNewRenderEvent)
        zope.component.event.objectEventNotify(event)

        return self.render(fs=fs, id=None)

    def add_contract(self, *args, **kwargs):
        self.request.model_class = dashboard.Contract
        self.request.model_name = dashboard.Contract.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_contract']
        return self.new()

    def all_customer_requests(self):
        session = DBSession()
        return self.request.filter_viewables(
                        session.query(dashboard.CustomerRequest)\
                               .filter(dashboard.CustomerRequest.project ==\
                                       self.context.get_instance()))

    @actions.action()
    def time_entries(self, *args, **kwargs):
        context = self.context.get_instance()
        page = self.get_page(collection=context.time_entries)
        pager = page.pager(**self.pager_args)
        return self.render(pager=pager, items=page)

    @actions.action()
    def tickets(self, *args, **kwargs):
        return self.render(customer_requests=self.all_customer_requests())

    @actions.action()
    def contracts(self, *args, **kwargs):
        context = self.context.get_instance()
        page = self.get_page(collection=context.contracts)
        pager = page.pager(**self.pager_args)
        return self.render(pager=pager, items=page)

    @actions.action()
    def customer_requests(self, *args, **kwargs):
        project = self.context.get_instance()
        params = Backlog(self.request).backlog(projects=[project])
        params['request'] = self.request
        params['context'] = self.request.context
        params['all_contracts'] = project.contracts_by_state()
        params['multiple_bgb'] = False
        tickets = project.get_number_of_tickets_per_cr()
        params['number_of_tickets_per_cr'] = tickets and tickets or {}
        return SkinObject('tekken')(**params)

    def update_cr_states(self):
        context = self.context.get_instance()
        n_cr = 0

        cr_ids = self.request.POST.getall('cr')
        new_state = self.request.POST.get('new_state')
        for cr in DBSession().query(dashboard.CustomerRequest).filter(dashboard.CustomerRequest.id.in_(cr_ids)):
            try:
                workflow = get_workflow(cr, cr.__class__.__name__)
                workflow.transition_to_state(cr, self.request, new_state, skip_same=True)
                n_cr += 1
            except WorkflowError as msg:
                self.request.add_message(msg, 'warning')

        self.request.add_message('%d customer requets updated.' % n_cr)
        raise exc.HTTPFound(location=self.request.fa_url('Project', context.id, 'customer_requests'))

    @actions.action('documentation')
    def documentation(self, *args, **kwargs):
        raise HTTPFound('%s/applications' % self.request.current_route_url())

    @actions.action('configuration')
    def configuration(self, *args, **kwargs):
        context = self.context.get_instance()
        return self.render(groups=context.groups)

    @actions.action('applications')
    def applications(self, *args, **kwargs):
        context = self.context.get_instance()
        return self.render(applications=context.applications)

    def toggle_favorite(self, *args, **kwargs):
        user = self.request.authenticated_user
        context = self.context.get_instance()
        if user not in context.favorite_users:
            context.favorite_users.append(user)
            self.request.add_message(u'Project %s has been marked as favorite.' % context.name, 'success')
        else:
            context.favorite_users.remove(user)
            self.request.add_message(u'Project %s has been unmarked as favorite.' % context.name, 'success')
        raise HTTPFound(self.request.fa_url('Project', context.id))

    @actions.action('listing')
    def datatable(self, **kwargs):
        result = super(ProjectModelView, self).datatable(**kwargs)

        fs = result['fs']
        fs.configure(pk=True, readonly=True)
        viewed_projects = set(fs.rows.items)

        columns = ['name', 'customer', 'manager', 'customer_requests', 'groups', 'activated', 'id']
        self.pick_columns(fs, columns)

        user = self.request.authenticated_user
        favorite_projects = set(p.id for p in set(user.favorite_projects).intersection(viewed_projects))

        return dict(result,
                    columns=columns,
                    js_favorite_projects=json.dumps(list(favorite_projects)))

    def list_customer_requests(self):
        project = self.context.get_instance()
        return [{'id': cr.id, 'name': str(cr)} for cr in project.customer_requests]



# -*- coding: utf-8 -*-
import itertools
import os
import re
import newrelic.agent

from zope.interface import implements
from pyramid.response import Response
from pyramid import decorator
from pyramid.request import Request
from pyramid.i18n import TranslationStringFactory
from pyramid.view import view_config
from pyramid.security import has_permission

from penelope.core import fanstatic_resources, messages
from penelope.core.interfaces import IBreadcrumbs, IPorRequest, ISidebar
from penelope.core.lib.htmlhelpers import get_application_link
from penelope.core.security import acl
from penelope.core.lib.helpers import chunks, unicodelower
from penelope.core.models import DBSession, Project, KanbanBoard
from penelope.core.models.interfaces import IProjectRelated


_ = TranslationStringFactory('por')


class PORRequest(Request):

    model_instance = None
    challenge_item = None
    charset = 'utf8'

    implements(IPorRequest)

    def __init__(self, *args, **kwargs):
        super(PORRequest, self).__init__(*args, **kwargs)
        self.charset = 'UTF-8'

    def _get_message_from_session(self):
        msgs = self.session.get(messages.STATUSMESSAGEKEY)
        if not msgs:
            msgs = messages.Messages()
            self.session[messages.STATUSMESSAGEKEY] = msgs
        return msgs

    def add_message(self, *args, **kwargs):
        """\
        Call this once or more to add messages to be displayed at the
        top of the web page.

        Examples:

        >>> request.add_message(u'A random warning message', 'warning')

        If no type is given it defaults to 'info'
        >>> request.add_message(u'A random info message')

        The arguments are:
            request:   instance of pyramid Request
            message:   a string, with the text message you want to show,
                        or a HTML fragment
            type:      optional, defaults to 'info'. The type determines how
                        the message will be rendered, as it is used to select
                        the CSS class for the message. Predefined types are:
                        'success' - for successful actions
                        'info' - for informational messages
                        'warning' - for warning messages
                        'error' - for messages about restricted access or errors.
        """
        msgs = self._get_message_from_session()
        msgs.add(*args, **kwargs)

    def show_messages(self):
        msgs = self._get_message_from_session()
        value = msgs.show()
        if self.response.status_int not in (301, 302, 304):
            del self.session[messages.STATUSMESSAGEKEY]
        return value

    @decorator.reify
    def title(self):
        return self.root.title

    @decorator.reify
    def newrelic(self):
        return newrelic.agent

    @decorator.reify
    def authenticated_user(self):
        user = None
        identity = self.environ.get('repoze.who.identity', None)
        if identity:
            user = identity.get('user', None)
        return user

    def isAuthenticated(self):
        return self.authenticated_user is not None

    def context_for_breadcrumbs(self, view):
        model = self.model_instance
        if model:
            return model
        else:
            return view

    def render_breadcrumbs(self, view):
        context = self.context_for_breadcrumbs(view)
        context.request = view.request
        breadcrumbs = IBreadcrumbs(context)
        return breadcrumbs.render(self)

    def render_sidebar(self, view):
        context = self.context_for_breadcrumbs(view)
        context.request = view.request
        sidebar = ISidebar(context)
        return sidebar.render(self)

    def fa_parent_url(self, context=None):
        if IProjectRelated.providedBy(self.model_instance):
            return self.fa_url('Project', self.model_instance.project.id)
        else:
            return self.fa_url(self.model_name)

    def active_topbar(self, paths):
        for path in paths:
            if path in self.path:
                return 'active'
        return ''

    def filter_viewables(self, items):
        if len([a for a in self.authenticated_user.roles if a.name == 'administrator']) == 1:
            # for administrator let's skip check 
            return items
        return [item for item in items if self.filter_viewable(item)]

    def filter_viewable(self, item):
        self.challenge_item = item
        if has_permission('view', item, self):
            self.challenge_item = None
            return item
        self.challenge_item = None

    def has_permission(self, permission, context):
        if not context:
            context = DefaultContext(self) #we always need a context for ACL to work
        self.challenge_item = context
        result = has_permission(permission, context, self)
        self.challenge_item = None
        return result


class DefaultContext(object):
    """
    Default context factory for TP views.
    """
    __acl__ = acl.DEFAULT_ACL

    def __init__(self, request):
        fanstatic_resources.dashboard.need()
        self.request = request

    def prettify_title(self, title):
        title = title.replace('_', ' ')
        return title

    @property
    def title(self):
        view = self.request.view_name or self.request.environ['PATH_INFO'].split('/')[-1]
        if view == '':
            title = 'Dashboard'
        elif view == 'add_entry':
            title = 'Smart add time entries'
        else:
            title = self.prettify_title(view).title()
        if not title:
            title = 'Penelope'
        return title


@view_config(context='pyramid.exceptions.NotFound', renderer='penelope.core:templates/notfound.pt')
def notfound_view(request):
    request.response.status_int = 404
    return {}


def group_by_customer(projects):
    return  [
                (x[0], list(x[1]))
                for x in itertools.groupby(projects, lambda x: x.customer)
            ]

@view_config(permission='view_home', route_name="favicon")
def favicon(request):
    here = os.path.dirname(__file__)
    icon = open(os.path.join(here, 'static', 'images', 'favicon.png'))
    return Response(content_type='image/x-icon', app_iter=icon)


@view_config(renderer='skin', permission='view_home')
def view_home(request):
    """
    Default home view
    """
    fanstatic_resources.dashboard_home.need()

    session = DBSession()
    user = request.authenticated_user
    projects = session.query(Project)

    my_projects = projects\
                  .filter(Project.users_favorite(user))\
                  .order_by(Project.customer_id).all()

    my_projects = request.filter_viewables(my_projects)

    active_projects = set(projects.filter(Project.active))
    active_projects = sorted(request.filter_viewables(active_projects.difference(my_projects)), key=unicodelower)

    listings = []
    listing_columns = 4
    listings.append({'title': 'Active projects',
                     'projgroups': tuple(chunks(tuple(active_projects), listing_columns)),})
    kanbanboards = user.kanban_boards

    return {
            'my_projects': my_projects,
            'kanbanboards': kanbanboards,
            'listings': listings,
            'can_see_kanbanboards':  request.has_permission('listing', KanbanBoard),
            'get_application_link': get_application_link,
            }


@view_config(renderer='skin', permission='view_home', route_name='navbar')
def view_navbar(request):
    """
    Navbar is rendered separately, and served to Trac via ajax.
    """
    trac = re.search('/trac/(?P<trac>[a-zA-z0-9\-_]+)',request.environ.get('HTTP_REFERER', ''))
    if trac:
        return {'trac':trac.group('trac')}
    else:
        return {}



# -*- coding: utf-8 -*-

def get_application_link(request, application):
    if application.application_type == 'trac':
        return application.api_uri

    return '%s/admin/Application/%s' % (request.application_url, application.id)



# -*- coding: utf-8 -*-

import collections
import json
import logging

import colander
from sqlalchemy.orm import lazyload
from sqlalchemy import distinct
from colander import SchemaNode

from deform import ValidationFailure
from deform.widget import SelectWidget
from pyramid.httpexceptions import HTTPForbidden
from pyramid.view import view_config

from penelope.core.models import DBSession, Project, TimeEntry, User, CustomerRequest, Contract, Customer
from penelope.core.models.tickets import ticket_store
from penelope.core import fanstatic_resources
from penelope.core.lib.helpers import ticket_url, total_seconds
from penelope.core.lib.widgets import SearchButton, PorInlineForm
from penelope.core.reports.queries import qry_active_projects, te_filter_by_customer_requests, filter_users_with_timeentries, te_filter_by_contracts
from penelope.core.reports.validators import validate_period
from penelope.core.reports.favourites import render_saved_query_form
from penelope.core.reports import fields

log = logging.getLogger(__name__)


class IDTree(dict):
    # XXX see if we can do without this silly tree

    def insert_entry(self, groupby, te_id, customer, project, request, user, date, **discard):
        """
        populates the tree with time-entry ids
        """
        node = self
        for idx, nodetype in enumerate(groupby, 1):
            if idx < len(groupby):
                default = {}
            else:
                default = []

            if nodetype == 'customer':
                node.setdefault(customer, default)
                node = node[customer]
            elif nodetype == 'project':
                node.setdefault(project, default)
                node = node[project]
            elif nodetype == 'request':
                node.setdefault(request, default)
                node = node[request]
            elif nodetype == 'user':
                node.setdefault(user, default)
                node = node[user]
            elif nodetype == 'date':
                node.setdefault(date, default)
                node = node[date]
            else:
                raise ValueError('unsupported groupby value?')

        node.append(te_id)


class AllEntriesReport(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request


    class AllEntriesSchema(colander.MappingSchema):
        customer_id = fields.customer_id.clone()
        project_id = fields.project_id.clone()
        date_from = fields.date_from.clone()
        date_to = fields.date_to.clone()
        users = fields.users.clone()
        contracts = fields.contracts.clone()
        customer_requests = fields.customer_requests.clone()

        groupbyfirst = SchemaNode(colander.String(),
                                  validator=colander.OneOf([
                                                            'project.request.user.date',
                                                            'project.request.date.user',
                                                            'project.user.request.date',
                                                            'project.user.date.request',
                                                            'project.date.user.request',
                                                            'user.project.request.date',
                                                            'user.project.date.request',
                                                            'user.date.project.request',
                                                            'date.user.project.request',
                                                            ]),
                                  widget=SelectWidget(values=[
                                          ('project.request.user.date', 'Project/Request/User/Date'),
                                          ('project.request.date.user', 'Project/Request/Date/User'),
                                          ('project.user.request.date', 'Project/User/Request/Date'),
                                          ('project.user.date.request', 'Project/User/Date/Request'),
                                          ('project.date.user.request', 'Project/Date/User/Request'),
                                          ('user.project.request.date', 'User/Project/Request/Date'),
                                          ('user.project.date.request', 'User/Project/Date/Request'),
                                          ('user.date.project.request', 'User/Date/Project/Request'),
                                          ('date.user.project.request', 'Date/User/Project/Request'),
                                      ]),
                                  default='project.request.user.date',
                                  missing='project.request.user.date',
                                  title=u'')



    def search(self, customer_id, project_id, date_from, date_to,
               users, customer_requests, groupbyfirst, contracts):

        # also search archived projects, if none are specified
        qry = DBSession.query(TimeEntry).join(TimeEntry.project).join(Project.customer).outerjoin(TimeEntry.author)
        qry = qry.options(lazyload(TimeEntry.project, TimeEntry.author, Project.customer))

        groupby = ['customer']

        groupby.extend(groupbyfirst.split('.'))

        if customer_id is not colander.null:
            qry = qry.filter(Project.customer_id == customer_id)
            groupby.remove('customer')

        if project_id is not colander.null:
            qry = qry.filter(TimeEntry.project_id == project_id)
            try:
                groupby.remove('customer')
            except ValueError:
                pass
            groupby.remove('project')

        if date_from is not colander.null:
            qry = qry.filter(TimeEntry.date>=date_from)

        if date_to is not colander.null:
            qry = qry.filter(TimeEntry.date<=date_to)

        if date_from == date_to and date_from is not colander.null:
            groupby.remove('date')

        if users:
            qry = qry.filter(TimeEntry.author_id.in_(users))

        qry = qry.filter(te_filter_by_customer_requests(customer_requests, request=self.request))

        qry = qry.filter(te_filter_by_contracts(contracts))

        rows = []

        id_tree = IDTree()

        time_entries = list(self.request.filter_viewables(qry))

        proj_tickets = collections.defaultdict(set)
        for te in time_entries:
            if te.ticket is None:
                continue
            proj_tickets[te.project_id].add(te.ticket)

        # projectsmap = {
        #    'project_id': {
        #         'ticket_id': 'customer_request_id',
        #         ...
        #    },
        #    ...
        # }

        projectsmap = {}
        for project_id, ticket_ids in proj_tickets.items():
            project = DBSession.query(Project).get(project_id)
            if not self.request.has_permission('reports_all_entries_for_project', project):
                continue
            projectsmap[project_id] = dict(ticket_store.get_requests_from_tickets(
                                            project, tuple(ticket_ids)))

        for te in time_entries:
            if not self.request.has_permission('reports_all_entries_for_project', te.project):
                continue

            customer_request = te.customer_request
            entry = {
                        'customer': te.project.customer.name.strip(),
                        'project': te.project.name.strip(),
                        'request': customer_request.name.strip(),
                        'user': te.author.fullname.strip(),
                        'date': te.date.strftime('%Y-%m-%d'),
                        'description': te.description,
                        'seconds': total_seconds(te.hours),
                        'time': 'ore',
                    }

            rows.append(entry)
            id_tree.insert_entry(groupby, te.id, **entry)

        return {
                'groupby': groupby,
                'rows': rows,
                'id_tree': id_tree,
                }


    @view_config(name='report_all_entries', route_name='reports', renderer='skin', permission='reports_all_entries')
    def __call__(self):
        fanstatic_resources.report_all_entries.need()

        schema = self.AllEntriesSchema(validator=validate_period).clone()

        if len([a for a in self.request.authenticated_user.roles if a.name == 'administrator']) == 1:
            projects = qry_active_projects()
        else:
            projects = [
                    project for project in self.request.filter_viewables(qry_active_projects())
                    if self.request.has_permission('reports_all_entries_for_project', project)
                    ]

        project_ids = [p.id for p in projects]
        customers = DBSession.query(distinct(Customer.id), Customer.id, Customer.name).join(Project).filter(Project.id.in_(project_ids)).order_by(Customer.name)
        users = DBSession.query(User).order_by(User.fullname)
        users = filter_users_with_timeentries(users)
        customer_requests = DBSession.query(CustomerRequest.id, CustomerRequest.name).order_by(CustomerRequest.name)
        contracts = DBSession.query(Contract.id, Contract.name).order_by(Contract.name)


        form = PorInlineForm(schema,
                             formid='all_entries',
                             method='GET',
                             buttons=[
                                 SearchButton(title=u'Search'),
                             ])

        form['customer_id'].widget.values = [('', '')] + [(str(c.id), c.name) for c in customers]
        # don't validate as it might be an archived customer
        form['project_id'].widget.values = [('', '')] + [(str(p.id), p.name) for p in projects]
        # don't validate as it might be an archived project

        form['users'].widget.values = [(str(u.id), u.fullname) for u in users]
        form['customer_requests'].widget.values = [(str(c.id), c.name) for c in customer_requests]
        form['contracts'].widget.values = [(str(c.id), c.name) for c in contracts]

        controls = self.request.GET.items()

        if not controls:
            # the form is empty
            return {
                    'form': form.render(),
                    'saved_query_form': render_saved_query_form(self.request),
                    'qs':'',
                    'has_results': False
                    }

        try:
            appstruct = form.validate(controls)
        except ValidationFailure as e:
            return {
                    'form': e.render(),
                    'saved_query_form': render_saved_query_form(self.request),
                    'qs': self.request.query_string,
                    'has_results': False
                    }

        entries_detail = self.search(**appstruct)

        col_headers = {
            'customer': u'Customer',
            'project': u'Project',
            'request': u'Request',
            'user': u'User',
            'date': u'Data',
        }

        columns = [
            { 'colvalue': 'time', 'groupbyrank': None, 'pivot': True, 'result': False },
        ]

        for idx, colname in enumerate(entries_detail['groupby']):
            columns.append({
                            'colvalue': colname,
                            'coltext': colname,
                            'header': col_headers[colname],
                            'groupbyrank': idx+1,
                            'pivot': False,
                            'result': False
                           })

        columns.append({'colvalue': 'description', 'coltext': 'description', 'header': 'description', 'pivot': False, 'result': False})
        columns.append({'colvalue': 'seconds', 'groupbyrank': None, 'pivot': False, 'result': True})

        sourcetable = {
                    'rows': entries_detail['rows'],
                    'columns': columns
                }

        return {
                'form': form.render(appstruct=appstruct),
                'saved_query_form': render_saved_query_form(self.request),
                'qs': self.request.query_string,
                'has_results': len(sourcetable['rows'])>0,
                'tpReport_oConf': json.dumps({
                                    'sourcetable': sourcetable,
                                    'id_tree': entries_detail['id_tree'],
                                    'groupby': entries_detail['groupby'],
                                }),
                }



    @view_config(name='all_entries_xls', route_name='reports', renderer='xls_report', permission='reports_all_entries')
    @view_config(name='all_entries_csv', route_name='reports', renderer='csv_report', permission='reports_all_entries')
    def all_entries_download(self):
        schema = self.AllEntriesSchema().clone()
        form = PorInlineForm(schema)
        controls = self.request.GET.items()
        appstruct = form.validate(controls)

        header = ['cliente', 'progetto', 'request', 'user', 'data', 'descrizione', 'ore']

        rows = [
                (
                    row['customer'],
                    row['project'],
                    row['request'],
                    row['user'],
                    row['date'],
                    row['description'],
                    row['seconds']/60.0/60.0,
                )
            for row in self.search(**appstruct)['rows']
            ]

        return {
                'header': header,
                'rows': rows
                }


@view_config(name='report_te_detail', route_name='reports', renderer='skin', permission='reports_all_entries')
def report_te_detail(context, request):
    ids = map(int, request.params['ids'].split(','))
    qry = DBSession.query(TimeEntry).filter(TimeEntry.id.in_(ids)).order_by(TimeEntry.id)
    projects = set(te.project for te in qry)
    for project in projects:
        if not request.has_permission('reports_all_entries_for_project', project):
            return HTTPForbidden()
    return {
            'time_entries': qry.all(),
            'ticket_url': ticket_url,
            }




from raven.middleware import Sentry
from raven.base import Client
from raven.utils.wsgi import get_current_url, get_headers, get_environ


class PenelopeSentry(Sentry):

    def cleanup_headers(self, headers):
        if 'Cookie' in headers:
            del headers['Cookie']
        return headers

    def handle_exception(self, environ):
        event_id = self.client.captureException(
            data={
                'sentry.interfaces.Http': {
                    'method': environ.get('REQUEST_METHOD'),
                    'url': get_current_url(environ, strip_querystring=True),
                    'query_string': environ.get('QUERY_STRING'),
                    'headers': self.cleanup_headers(dict(get_headers(environ))),
                    'env': dict(get_environ(environ)),
                }
            },
        )
        return event_id


def sentry_filter_factory(app, global_conf, **kwargs):
    client = Client(**kwargs)
    return PenelopeSentry(app, client)



# -*- coding: utf-8 -*-

"""
FormAlchemy widgets for penelope.core
"""

from formalchemy.fields import FieldRenderer
from formalchemy import helpers as h


class BigTextAreaFieldRenderer(FieldRenderer):
    """
    A textarea with a bigger default size
    """
    def render(self, size=(60, 8), **kwargs):
        if isinstance(size, tuple):
            kwargs['size'] = 'x'.join(map(str, size))
        return h.text_area(self.name, content=self.value, **kwargs)





# This Python file uses the following encoding: utf-8
import colander

from plone.i18n.normalizer import idnormalizer
from colander import SchemaNode
from deform import ValidationFailure
from deform.widget import CheckboxWidget, TextInputWidget, SequenceWidget
from deform_bootstrap.widget import ChosenSingleWidget, ChosenMultipleWidget
from pyramid.renderers import get_renderer
from pyramid import httpexceptions as exc
from pyramid.i18n import TranslationStringFactory

from penelope.core.models import Project, Group, DBSession, User, CustomerRequest
from penelope.core.models.dashboard import ApplicationACL, Contract
from penelope.core.models.dashboard import Trac, Role, GoogleDoc, Estimation
from penelope.core.lib.widgets import SubmitButton, ResetButton, WizardForm
from penelope.core.fanstatic_resources import wizard as wizard_fanstatic
from penelope.core import PROJECT_ID_BLACKLIST
from penelope.core.notifications import notify_user_with_welcoming_mail

_ = TranslationStringFactory('penelope')


PM_TICKETS = [
 (_(u'Riesami e Verifiche degli Elementi in Ingresso al progetto'),
 _(u"""Questo ticket serve sia per i RIESAMI sia per le VERIFICHE del progetto.
Il RIESAME risponde alle domande HO TUTTI GLI ELEMENTI PER LAVORARE?
STO LAVORANDO SECONDO I REQUISITI DEL CLIENTE? In questo ticket viene
registrato l'avvenuto esame degli elementi iniziali (ad esempio: il
capitolato tecnico) e di tutti gli elementi emersi in corso d'opera (ad
esempio: documenti integrativi). La VERIFICA risponde alle domande HO
FATTO TUTTO QUELLO CHE C’ERA DA FARE? SONO PRONTO AL COLLAUDO? Cioè, per
capire se la produzione è conclusa, verifico che ciò che ho prodotto sia
compatibile con gli input della progettazione. POSSIBILMENTE, CI DOVREBBE
ESSERE UN COMMENT CON L'ELENCO DEI DOCUMENTI CONSULTATI, E UN COMMENT PER
OGNI FASE DI VERIFICA. TECNICAMENTE, QUESTO TICKET ANDREBBE CHIUSO PRIMA
DI PROCEDERE ALLA VALIDAZIONE DEL PROGETTO."""),
  'verification'),

 (_(u'Validazione del progetto'),
 _(u"""Questo ticket serve per la VALIDAZIONE, step finale del ciclo
produttivo, prima della consegna, ovvero la registrazione del fatto che
HO COLLAUDATO IL PRODOTTO IN UN CASO D’USO REALE E CI METTO LA FIRMA! La
validazione delle singole funzionalità avviene a livello di ticket. Questo
ticket viene completato (chiuso) per registrare l'avvenuta validazione del
progetto nel suo assieme. """),
 'verification'),

 (_(u'project management'), u'', 'task'),

 (_(u'incontri con il cliente'), u'', 'task'),

 (_(u'Punto interno con/tra sviluppatori'), u'', 'task'),

 (_(u'Riesame documentazione di progetto (verification)'), u'', 'verification'),

 (_(u'Gestione progetto su Penelope'), u'','task')]


class Definition(colander.Schema):

    def check_project_id(value):
        project_id = idnormalizer.normalize(value)
        project = DBSession().query(Project).get(project_id)
        if value.lower() in PROJECT_ID_BLACKLIST or project:
            return _('${project_id} is a restricted name or already exists! '
                     'Please choose another project name or provide unique ID!',
                     mapping={'project_id': value})
        else:
            return True

    project_name = SchemaNode(typ=colander.String(),
                              widget=TextInputWidget(css_class='input-xlarge',
                                                     placeholder=u'Project name'),
                              missing=colander.required,
                              validator=colander.All(colander.Length(max=20),
                                                     colander.Function(check_project_id)),
                              title=u'')
    trac_name = SchemaNode(typ=colander.String(),
                           widget=TextInputWidget(css_class='input-xxlarge',
                                                  placeholder=u"Short name. "\
                                          "It will appear in email's subject"),
                           missing=None,
                           title=u'')


class GoogleDocsSchema(colander.SequenceSchema):
    class GoogleDocSchema(colander.Schema):
        name = SchemaNode(typ=colander.String(),
                          widget=TextInputWidget(css_class='input-xlarge',
                                                 validator=colander.
                                                               Length(max=20),
                                                 placeholder=u'Enter google '
                                                             'doc name'),
                          missing=colander.required,
                          title=u'')
        uri = SchemaNode(typ=colander.String(),
                         widget=TextInputWidget(css_class='input-xxlarge',
                                                placeholder=u'Paste your '
                                                'google docs folder'),
                         missing=colander.required,
                         title=u'')
        share_with_customer = SchemaNode(typ=colander.Boolean(),
                                         widget=CheckboxWidget(),
                                         missing=None,
                                         title=u'Share with the customer')

    google_doc = GoogleDocSchema(title='')


class UsersSchema(colander.SequenceSchema):
    class UserSchema(colander.Schema):
        usernames = SchemaNode(colander.Set(),
                               widget=ChosenMultipleWidget(placeholder=
                                                           u'Select people'),
                               missing=colander.required,
                               title=u'')
        role = SchemaNode(typ=colander.String(),
                          widget=ChosenSingleWidget(),
                          missing=colander.required,
                          title=u'role')

    user = UserSchema(title='')


class NewUsersSchema(colander.SequenceSchema):
    class NewUserSchema(colander.Schema):
        def unusedEmail(value):
            user = DBSession.query(User.id).filter(User.email == value).first()
            if user:
                return "email '%s' is already associated to another user" % \
                                                                         value
            else:
                return True

        fullname = SchemaNode(typ=colander.String(),
                              widget=TextInputWidget(placeholder=u'Fullname'),
                              missing=colander.required,
                              title=u'')
        email = SchemaNode(typ=colander.String(),
                           widget=TextInputWidget(placeholder=u'E-mail'),
                           missing=colander.required,
                           validator=colander.Function(unusedEmail, ''),
                           title=u'')
        role = SchemaNode(typ=colander.String(),
                          widget=ChosenSingleWidget(),
                          missing=colander.required,
                          title=u'Role')
        send_email_howto = SchemaNode(typ=colander.Boolean(),
                                      widget=CheckboxWidget(),
                                      missing=None,
                                      title=u'Send e-mail')

    new_user = NewUserSchema(title='')


class Milestones(colander.SequenceSchema):
    class Milestone(colander.Schema):
        title = SchemaNode(typ=colander.String(),
                           widget=TextInputWidget(placeholder=u'Title'),
                           missing=colander.required,
                           title=u'')
        due_date = SchemaNode(typ=colander.Date(),
                              missing=colander.required,
                              title=u'Due date')

    milestone = Milestone(title='')


class CustomerRequests(colander.SequenceSchema):
    class CustomerRequest(colander.Schema):
          """
              #BBB specify that junior & co wants days
          """
          title = SchemaNode(typ=colander.String(),
                             widget=TextInputWidget(placeholder=u'Title, '
                                                    'the customer wants...'),
                             missing=colander.required,
                             title=u'')
          junior = SchemaNode(typ=colander.Decimal(),
                              widget=TextInputWidget(css_class='input-mini',
                                                     placeholder=u'Junior'),
                              missing=None,
                              title=u'')
          senior = SchemaNode(typ=colander.Decimal(),
                              widget=TextInputWidget(css_class='input-mini',
                                                     placeholder=u'Senior'),
                              missing=None,
                              title=u'')
          graphic = SchemaNode(typ=colander.Decimal(),
                               widget=TextInputWidget(css_class='input-mini',
                                                      placeholder=u'Graphic'
                                                      ),
                               missing=None,
                               title=u'')
          pm = SchemaNode(typ=colander.Decimal(),
                          widget=TextInputWidget(css_class='input-mini',
                                                 placeholder=u'PM'),
                          missing=None,
                          title=u'')
          architect = SchemaNode(typ=colander.Decimal(),
                                 widget=TextInputWidget(css_class='input-mini',
                                                        placeholder=u'Arch.'),
                                 missing=None,
                                 title=u'')
          tester = SchemaNode(typ=colander.Decimal(),
                              widget=TextInputWidget(css_class='input-mini',
                                                     placeholder=u'Tester'),
                              missing=None,
                              title=u'')
          ticket = SchemaNode(typ=colander.Boolean(),
                              widget=CheckboxWidget(),
                              missing=None,
                              title=u'Create related ticket')

    customer_request = CustomerRequest(title='')


class CRperContracts(colander.MappingSchema):
    class Contract(colander.Schema):
          name = SchemaNode(typ=colander.String(),
                             widget=TextInputWidget(placeholder=u'Contract name'),
                             missing=colander.required,
                             title=u'')
          contract_number = SchemaNode(typ=colander.String(),
                             widget=TextInputWidget(placeholder=u'Contract '
                                                    'number'),
                             missing=None,
                             title=u'')
          days = SchemaNode(typ=colander.Decimal(),
                              widget=TextInputWidget(css_class='input-mini',
                                                     placeholder=u'Days'),
                              missing=None,
                              title=u'')
          amount = SchemaNode(typ=colander.Decimal(),
                              widget=TextInputWidget(css_class='input-mini',
                                                     placeholder=u'amount in EUR'),
                              missing=None,
                              title=u'')
          start_date = SchemaNode(typ=colander.Date(),
                              missing=None,
                              title=u'Start date')
          end_date = SchemaNode(typ=colander.Date(),
                              missing=None,
                              title=u'End date')

    customer_requests = CustomerRequests(missing=None)
    contract = Contract(name='',)


class WizardSchema(colander.Schema):
    project = Definition()
    google_docs = GoogleDocsSchema(missing=None)
    users = UsersSchema(missing=None)
    new_users = NewUsersSchema(missing=None)
    milestones = Milestones(missing=None)
    contracts = CRperContracts(missing=None)


class Wizard(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def render(self):
        result = {}
        result['main_template'] = get_renderer(
                'penelope.core:skins/main_template.pt').implementation()
        result['main'] = get_renderer(
                'penelope.core.forms:templates/master.pt').implementation()

        schema = WizardSchema().clone()
        wizard_fanstatic.need()
        form = WizardForm(schema,
                          formid='wizard',
                          method='POST',
                          buttons=[
                                 SubmitButton(title=u'Submit'),
                                 ResetButton(title=u'Reset'),
                          ])
        form['new_users'].widget = SequenceWidget()
        form['users'].widget = SequenceWidget(min_len=1)

        users = DBSession.query(User).order_by(User.fullname)
        form['users']['user']['usernames'].widget.values = [('', '')] + \
                                      [(str(u.id), u.fullname) for u in users]

        roles = DBSession.query(Role).order_by(Role.name)
        form['users']['user']['role'].widget.values = [('', '')] + \
                                 [(str(role.id), role.name) for role in roles]
        form['new_users']['new_user']['role'].widget.values = [('', '')] + \
                [(str(role.id), role.name) for role in roles]

        form['milestones'].widget = SequenceWidget(min_len=1)
        form['contracts'].title = ''
        form['contracts']['customer_requests'].widget = SequenceWidget(min_len=3)

        controls = self.request.POST.items()
        if controls != []:
            try:
                appstruct = form.validate(controls)
                self.handle_save(form, appstruct)
            except ValidationFailure as e:
                result['form'] = e.render()
                return result

        appstruct = {}
        appstruct['contracts'] ={'customer_requests': []}
        appstruct['contracts']['customer_requests'].append({'ticket': True,
                                                            'title': u'Project management'})
        appstruct['contracts']['customer_requests'].append({'ticket': True,
                                                            'title': u'Analisi'})
        appstruct['contracts']['customer_requests'].append({'ticket': True,
                                                            'title': u'Supporto'})
        result['form'] = form.render(appstruct=appstruct)
        return result

    def handle_save(self, form, appstruct):
        """ The main handle method for the wizard. """
        customer = self.context.get_instance()

        # create new users
        recipients = []
        groups = {}
        for newuser in appstruct['new_users']:
            user = User(fullname=newuser['fullname'], email=newuser['email'])
            if not newuser['role'] in groups:
                groups[newuser['role']] = []
            groups[newuser['role']].append(user)
            if newuser['send_email_howto']:
                recipients.append(newuser['email'])

        for recipient in recipients:
            notify_user_with_welcoming_mail(recipient)

        #create project and set manager
        manager = self.request.authenticated_user
        project = Project(name=appstruct['project']['project_name'],
                          manager=manager)

        #set groups
        for g in appstruct['users']:
            if not g['role'] in groups:
                groups[g['role']] = []
            for u in g['usernames']:
                user = DBSession.query(User).get(u)
                groups[g['role']].append(user)

        for rolename, users in groups.items():
            role = DBSession.query(Role).filter(Role.name == rolename).one()
            group = Group(roles=[role, ], users=users)
            project.add_group(group)

        #create contract with cr
        crs = appstruct['contracts']['customer_requests']
        co =  appstruct['contracts']['contract']
        contract = Contract(**co)
        contract.project_id = project.id

        #create CR
        tickets = []
        for cr in crs:
            customer_request = CustomerRequest(name=cr['title'])
            person_types = {
                'junior': 'Junior',
                'senior': 'Senior',
                'graphic': 'Graphic',
                'pm': 'Project manager',
                'architect': 'Architect',
                'tester': 'Tester'
            }
            for key, value in person_types.items():
                if cr[key]:
                    Estimation(person_type=value,
                              days=cr[key],
                              customer_request=customer_request)
                    customer_request.workflow_state = 'estimated'
            project.add_customer_request(customer_request)
            customer_request.contract = contract

            if cr['title'] == 'Project management':
                for summary, description, ticket_type in PM_TICKETS:
                    tickets += [{ 'summary': summary,
                                  'description': description,
                                  'customerrequest': customer_request,
                                  'reporter': manager.email,
                                  'type': ticket_type,
                                  'priority': 'major',
                                  'sensitive': '1',
                                  'milestone': 'Backlog',
                                  'owner': manager.email}]

            elif cr['ticket']:
                tickets += [{'summary': cr['title'],
                            'customerrequest': customer_request,
                            'reporter': manager.email,
                            'type': 'task',
                            'priority': 'major',
                            'milestone': 'Backlog',
                            'owner': manager.email}]

        #create google docs/folders
        for app_definition in appstruct['google_docs']:
            app = GoogleDoc(name=app_definition['name'],
                            api_uri=app_definition['uri'])
            if app_definition['share_with_customer']:
                acl = ApplicationACL(role_id='customer',
                                     permission_name='view')
                acl.project = app
            project.add_application(app)

        #create trac
        trac = Trac(name="Trac for %s" % appstruct['project']['project_name'])
        trac.settings = self.request.registry.settings # an ugly patch to pass 
                                                       # the missing settings 
        trac.milestones = appstruct['milestones']
        trac.tickets = tickets
        if appstruct['project']['trac_name']:
            trac.project_name = appstruct['project']['trac_name']
        else:
            trac.project_name = appstruct['project']['project_name']
        project.add_application(trac)

        customer.add_project(project)
        raise exc.HTTPFound(location=self.request.fa_url('Customer',
                                                         customer.id))



import gspread
import getpass
import os
import sys
import transaction
import beaker

from sqlalchemy import engine_from_config
from pyramid.paster import setup_logging, bootstrap

from penelope.core.models.dbsession import DBSession
from penelope.core.models import Base
from penelope.core.models.dashboard import Contract, CustomerRequest
from penelope.core.models.tp import TimeEntry


beaker.cache.cache_regions.update(dict(calculate_matrix={'key_length':''}))


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri> [google_user] [google_spreadsheet key]\n'
          '(example: "%s development.ini")' % (cmd, cmd)) 
    sys.exit(1)


def split_dict_equally(input_dict, chunks=10):
    "Splits dict by keys. Returns a list of dictionaries."
    # prep with empty dicts
    return_list = [dict() for idx in xrange(chunks)]
    idx = 0
    for k,v in input_dict.iteritems():
        return_list[idx][k] = v
        if idx < chunks-1:  # indexes start at 0
            idx += 1
        else:
            idx = 0
    return return_list


def update_time_entries():
    session = DBSession()

    tracs = {}
    time_entries = session.query(TimeEntry)
    for tp in time_entries:
        trac = list(tp.project.tracs)[0]
        tracs.setdefault(trac.trac_name, [])
        if not tp.ticket:
            print "TimeEntry %s has no ticket" % tp.id
            continue
        tracs[trac.trac_name].append((tp.id, tp.ticket))

    ticket_vs_crs = """SELECT value AS cr_id, '%(tp_id)s' AS tp_id FROM "trac_%(trac)s".ticket_custom WHERE name='customerrequest' AND ticket=%(ticket)s"""
    splitted_tracs = split_dict_equally(tracs)

    for split in splitted_tracs:
        queries = []
        for trac_id, opts in split.items():
            for opt in opts:
                queries.append(ticket_vs_crs % {'trac': trac_id,
                                                'tp_id': opt[0],
                                                'ticket': opt[1]})
        sql = '\nUNION '.join(queries)
        sql += ';'
        for trac in DBSession().execute(sql).fetchall():
            cr = session.query(CustomerRequest).get(trac.cr_id)
            if not cr:
                continue
            contract = cr.contract
            if not contract:
                continue
            tp = session.query(TimeEntry).get(trac.tp_id)
            if not tp:
                continue
            tp.contract_id = contract.id

def map_state(state):
    if state == 'A':
        return 'active'
    if state == 'B':
        return 'draft'
    if state == 'D':
        return 'done'
    return 'draft'


def main(argv=sys.argv):
    if len(argv) < 2:
        usage(argv)

    config_uri = argv[1]
    setup_logging(config_uri)
    env = bootstrap('%s#dashboard'% config_uri)
    settings = env.get('registry').settings
    engine = engine_from_config(settings, 'sa.dashboard.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine

    google_opts = argv[2:]
    try:
        google_user = google_opts.pop(0)
    except IndexError:
        google_user = raw_input('Google Spreadsheet User: ')
    try:
        spreadsheet_key = google_opts.pop(0)
    except IndexError:
        spreadsheet_key = raw_input('Google Spreadsheet ID: ')
    google_password = getpass.getpass("Google Spreadsheet Password: ")

    gc = gspread.login(google_user, google_password)
    sht = gc.open_by_key(spreadsheet_key)
    worksheet = sht.get_worksheet(0)
    crs = worksheet.get_all_records()

    contracts = {}
    for row in crs:
        if not row['titolocommessa']:
            continue
        contract_uid = '%s_%s_%s' % (row['project_name'],
                                     row['titolocommessa'],
                                     row['customer_id'])
        contracts.setdefault(contract_uid, {'crs': []})
        contracts[contract_uid]['titolocommessa'] = row['titolocommessa']
        contracts[contract_uid]['nrcontratto'] = row['nrcontratto']
        contracts[contract_uid]['gg'] = row['gg'] or 0
        contracts[contract_uid]['amount'] = row['amount'] or 0
        contracts[contract_uid]['crs'].append(row['cr_id'])
        contracts[contract_uid]['stato'] = map_state(row['stato'])

    # now we have a structure:
    # contracts['ContractUID'] = {'crs': ['customer_request_id_1',
    #                                     'customer_request_id_2'],
    #                             'gg': '12'}

    with transaction.manager:
        session = DBSession()
        for contract_uid, opts in contracts.items():
            crs = [session.query(CustomerRequest).get(a) for a in opts['crs']]
            crs = [a for a in crs if a]
            if not crs:
                continue
            contract = crs[0].contract
            if not contract:
                contract = Contract(name=opts['titolocommessa'])
            contract.days = opts['gg']
            contract.amount = opts['amount']
            contract.contract_number = opts['nrcontratto']
            contract.workflow_state = opts['stato']
            for cr in crs:
                if not cr:
                    continue
                cr.contract = contract
                contract.project_id = cr.project_id

        update_time_entries()



# -*- coding: utf-8 -*-

TEMPLATE = """<div class="alert alert-%(type)s fade in"><a class="close" href="#" data-dismiss="alert">x</a><p>%(message)s</p></div>"""
STATUSMESSAGEKEY = 'PENELOPE_MESSAGE'


class Message(object):
    """A single status message."""

    def __init__(self, message, type):
        self.message = message
        self.type = type


class Messages(list):
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self.template = TEMPLATE

    def show(self):
        return u''.join([self.template % a.__dict__ for a in self])

    def add(self, text, type='info'):
        message = Message(text, type)
        self.append(message)



# -*- coding: utf-8 -*-

import datetime

import sqlalchemy as sa

from pyramid.view import view_config

from penelope.core.models import DBSession
from penelope.core.models.tp import TimeEntry, timedelta_as_human_str
from penelope.core import views, fanstatic_resources


class TPContext(views.DefaultContext):
    """
    Default context factory for TP views.
    """


@view_config(name='add_entry', renderer='skin', permission='add_entry')
def add_entry(context, request):
    fanstatic_resources.add_entry.need()
    return {
            'context': context,
            'request': request,
            }


@view_config(name='latest_entries', renderer='skin')
def latest_entries(context, request):
    """
    Returns an HTML fragment with tables of the latest time entries
    """

    qry = DBSession.query(TimeEntry)

    current_user = request.environ.get('repoze.who.identity')['user']
    qry = qry.filter(TimeEntry.author_id==current_user.id)

    qry = qry.order_by(sa.desc(TimeEntry.date), sa.desc(TimeEntry.start), sa.desc(TimeEntry.creation_date))
    time_entries_today = qry.filter(TimeEntry.date==datetime.date.today()).all()
    today_total = timedelta_as_human_str(sum([a.hours for a in time_entries_today], datetime.timedelta()))

    latest_limit = 20
    time_entries_latest = qry.limit(latest_limit)

    return {
            'context': context,
            'request': request,
            'time_entries_today': time_entries_today,
            'today_total': today_total,
            'time_entries_latest': time_entries_latest,
            'time_entries_latest_limit': latest_limit,
            'report_my_from_date': datetime.date.today() - datetime.timedelta(days=6),
            'today': datetime.date.today(),
            }



# -*- coding: utf-8 -*-

import gdata.docs.data
import logging
import transaction
import re
import zope.component
from copy import deepcopy

from sqlalchemy import func
from gdata.client import RequestError
from pyramid_skins import SkinObject
from pyramid.renderers import get_renderer
from pyramid.exceptions import Forbidden
from pyramid.httpexceptions import HTTPFound
from pyramid.threadlocal import get_current_registry
from gdata.calendar.client import CalendarEventQuery
from dateutil.parser import parse
from datetime import timedelta

from penelope.core.sidebar import SidebarRenderer, HeaderSidebarAction, SidebarAction
from penelope.core.interfaces import ISidebar
from penelope.core.gdata import IIterationView
from penelope.core.gdata.utils import documents, spreadsheets, calendar, get_working_days
from penelope.core.models import Project, DBSession, User, GlobalConfig, Customer, Role


logging.basicConfig()
log = logging.getLogger(__file__)


def clear_token(context, request):
    identity = request.environ.get('repoze.who.identity')
    if not identity:
        raise Forbidden

    user = identity.get('user')
    if not user:
        raise Forbidden

    user.gdata_auth_token = None
    next_url = '%s/admin/User/%s/user_tokens' % (request.application_url, request.authenticated_user.id)
    transaction.commit()
    raise HTTPFound(next_url)


@documents.auth_required
def add_token(context, request):
    next_url = '%s/admin/User/%s/user_tokens' % (request.application_url, request.authenticated_user.id)
    raise HTTPFound(next_url)


def get_cell_values(request, query):
    docid = request.params.get('docid')
    docid = docid.split(':')[1]

    cells = request.gservice['SpreadsheetsService'].GetCellsFeed(docid, query=query)
    return [a.content.text for a in cells.entry]


@spreadsheets.auth_required
def activate_iteration(context, request):
    docid = request.params.get('docid')
    if not docid:
        return view_iterations(context, request, validation_error=u'Missing document_id')

    #query = gdata.spreadsheet.service.CellQuery()

    #first take project names
    #query['min-col'] = '3'
    #query['max-col'] = '3'
    #query['min-row'] = '5'
    #cr_raw = get_cell_values(request, query)

    session = DBSession()

    #deactivate all CR
    #for cr in session.query(CustomerRequest):
    #    cr.active = False

    #activate selected CR
    #cr_ids = set([item for sublist in [a.split(',') for a in cr_raw] for item in sublist])
    #crs = session.query(CustomerRequest).filter(CustomerRequest.id.in_(cr_ids))
    #for cr in crs:
    #    cr.active = True

    gc = session.query(GlobalConfig).get(1)
    gc.active_iteration_url = docid

    return manage_iterations(context,request)

@documents.auth_graceful
def get_google_document(context, request, **kwargs):

    is_folder = re.compile(r'folders/(.*)$').search(kwargs['uri'])
    is_document = re.compile(r'/([a-z]*)/d/(.*)/.*$').search(kwargs['uri'])
    is_spreadsheet = re.compile(r'/([a-z]*)/ccc\?key=([0-9a-zA-Z]*)').search(kwargs['uri'])
    params = {}
    status = None
    folder = None
    document = None

    if is_spreadsheet:
        is_document = is_spreadsheet

    if is_folder:
        folder_id = is_folder.groups()[0]
        try:
            folder = request.gclient['DocsClient'].get_doclist(uri='/feeds/default/private/full/%s/contents/' % folder_id).entry
        except (ValueError, RequestError), e:
            status = u'''Something is not working correctly.
                         We cannot open Google Collection you have provided.
                         Please check if %s is correct.
                         Full exception: %s''' % (kwargs['uri'], e)
        params.update(folder=folder, status=status)

    elif is_document:
        document_type, document_id = is_document.groups()
        try:
            document = request.gclient['DocsClient'].get_doc('%s:%s' % (document_type, document_id))
        except (ValueError, RequestError), e:
            status = u'''Something is not working correctly.
                         We cannot open Google Doc url you have provided.
                         Please check if %s is correct.
                         Full exception: %s''' % (kwargs['uri'], e)
        params.update(document=document,status=status)

    else:
        params.update(status='''Something is not working correctly. We cannot open Google Apps url you have provided.
        Please check if %s is correct''' % kwargs['uri'])

    return params


@documents.auth_graceful
def manage_iterations(context, request, **params):
    docs = []
    folder_iteration = get_iteration_folder(request)
    if folder_iteration:
        log.info("Iteration folder found: %s" % folder_iteration.resource_id.text)
        docs = request.gclient['DocsClient'].get_doclist(
                uri='/feeds/default/private/full/%s/contents/-/spreadsheet' % \
                                      folder_iteration.resource_id.text).entry
    settings = get_current_registry().settings
    domain = settings.get('penelope.core.google_domain')
    params.update({'context':context,
                   'docs': docs,
                   'iteration_folder': 'https://docs.google.com/a/%s/#%s' % (
                       domain,
                       folder_iteration.resource_id.text.replace('folder:','folders/')),
                   'request':request})

    return SkinObject('manage_iterations')(**params)


@documents.auth_graceful
def view_iterations(context, request, **params):
    session = DBSession()
    gc = session.query(GlobalConfig).get(1)
    docid = gc.active_iteration_url
    folder_iteration = get_iteration_folder(request)

    params.update({'context':context,
                   'doc_url': None,
                   'request':request})

    if folder_iteration and docid:
        log.info("Iteration folder found: %s" % folder_iteration.resource_id.text)
        document = request.gclient['DocsClient'].get_doc(docid)
        params.update({'doc_url': '%s&rm=minimal' % document.get_html_link().href})

    return SkinObject('view_iterations')(**params)


@spreadsheets.auth_required
def generate_spreadsheet(context, request):
    date_start = request.params.get('start')
    date_end = request.params.get('end')
    session = DBSession()

    client = request.gclient['DocsClient']
    service = request.gservice['SpreadsheetsService']
    iteration_folder = get_iteration_folder(request)
    settings = get_current_registry().settings

    if not date_start or not date_end:
        params = {'validation_error': 'Please select date range'}
        return manage_iterations(context, request, **params)

    if not iteration_folder:
        params = {'validation_error':
                       'Iteration folder is missing. Please create folder in google docs with title: %s' % \
                       settings.get('penelope.core.iteration_folder')}
        return manage_iterations(context, request, **params)

    users = session.query(User.email, User.fullname)\
                   .group_by(User.email, User.fullname)\
                   .join(User.roles).filter(Role.id.ilike('%developer%'))\
                   .order_by(func.substring(User.fullname, '([^[:space:]]+)(?:,|$)'))

    projects = session.query(Project).join(Customer)\
                      .filter(Project.active)\
                      .order_by(Customer.name, Project.name)


    dockey = settings.get('penelope.core.iteration_template')
    resourceid = 'document%%3A%s' % dockey
    template = client.GetDoc(resourceid)
    entry = client.copy(template, 'Iteration from %s to %s' % (date_start,
                                                               date_end))
    client.move(entry, iteration_folder)

    sp_id = entry.resource_id.text.split(':')[1]
    wk = service.GetWorksheetsFeed(sp_id).entry[0]
    wk.title.text = '%s_%s' % (date_start, date_end)
    wk = service.UpdateWorksheet(wk)
    wk_id = wk.id.text.split('/')[-1]

    query = gdata.spreadsheet.service.CellQuery()
    query['return-empty'] = 'true'
    query['min-col'] = '1'
    query['min-row'] = '1'
    query['max-col'] = '19'
    query['max-row'] = '106'
    cells = service.GetCellsFeed(sp_id, wksht_id=wk_id, query=query)
    batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()

    def update_cell(cells, values):
        for cell,value in zip(cells,values):
            cell.cell.inputValue = str(value)
            batchRequest.AddUpdate(cell)

    update_cell([cells.entry[19]],['Iteration from %s to %s' % (date_start, date_end)])
    #Add user fullnames
    update_cell(cells.entry[5:18], [a.fullname for a in users])
    #Add user working days
    dae = _get_calendars_events(users, request)
    dr = [parse(date_start), parse(date_end)]
    update_cell(cells.entry[24:37], [get_working_days(dr,dict(dae).get(u.email,[])) for u in users])
    #Add project names
    update_cell(cells.entry[76::19], [str(a) for a in projects])
    #Add project managers
    update_cell(cells.entry[77::19], [a.manager for a in projects])

    service.ExecuteBatch(batchRequest, cells.GetBatchLink().href)
    return manage_iterations(context, request)


@calendar.auth_required
def _get_calendars_events(users, request):

    """
    It retrieves all avalable calendars from the current user, this means also 'Festivita italiane' and 'Compleanni ed eventi'.
    The porpouse is to get all the events from each calendar with by keyword (in this example is 'holidays@google.com').
    and from each event get the date range value.

    return would be something like this:
    [
     ['user1@google.com',
        [[datetime.datetime(2011, 8, 24, 0, 0), datetime.datetime(2011, 8, 24, 23, 59)],
        [datetime.datetime(2011, 8, 23, 0, 0), datetime.datetime(2011, 8, 23, 23, 59)],
        [datetime.datetime(2011, 8, 23, 0, 0), datetime.datetime(2011, 8, 23, 23, 59)],
        [datetime.datetime(2011, 8, 22, 0, 0), datetime.datetime(2011, 8, 22, 23, 59)]]],
     ['user2@google.com',
        [[datetime.datetime(2011, 8, 24, 0, 0), datetime.datetime(2011, 8, 24, 23, 59)]]],
     ['%23contacts@group.v.calendar.google.com', []],
     ['it.italian%23holiday@group.v.calendar.google.com', []],
     ['user3@google.com', []]
    ]
    """
    result = []
    client = request.gclient['CalendarClient']

    # get all calendars
    query_holidays = CalendarEventQuery()
    query_holidays.start_min = request.params.get('start')
    query_holidays.start_max = request.params.get('end')

    cal_holidays_ranges = []
    try:
        italian_holidays = client.GetCalendarEventFeed(
                uri='https://www.google.com/calendar/feeds/en.italian%23holiday%40group.v.calendar.google.com/private/full',
                q=query_holidays)
        for holiday in italian_holidays.entry:
            s = parse(holiday.when[0].start)
            e = parse(holiday.when[0].end)
            cal_holidays_ranges.append([s, e-timedelta(minutes=1)])
    except RequestError: # gracefully ignore request errors
        pass

    settings = get_current_registry().settings
    attendees = settings.get('penelope.core.vacancy_email')
    query = CalendarEventQuery(text_query = attendees)
    query.start_min = request.params.get('start')
    query.start_max = request.params.get('end')

    for user in users:
        username = user.email
        feed_uri = client.GetCalendarEventFeedUri(calendar=username, visibility='private', projection='full')
        cal_events_ranges = deepcopy(cal_holidays_ranges)

        # get the event feed using the feed_uri and the query params in order to get only those with 'holidays@google.com'
        try:
            events_feed = client.GetCalendarEventFeed(uri=feed_uri, q=query)
            for an_event in events_feed.entry:
                if not an_event.when:
                    continue
                s = parse(an_event.when[0].start)
                e = parse(an_event.when[0].end)
                cal_events_ranges.append([s, e-timedelta(minutes=1)])
        except RequestError: # gracefully ignore request errors
            pass
        result.append([username,cal_events_ranges])
    return result


def get_iteration_folder(request):
    settings = get_current_registry().settings
    folders = request.gclient['DocsClient'].get_doclist(
                uri='/feeds/default/private/full?title=%s&category=folder' % \
                               settings.get('penelope.core.iteration_folder'))
    for folder in folders.entry:
        return folder


class IterationSidebarRenderer(SidebarRenderer):

    def render(self, request):
        self.actions.append(HeaderSidebarAction('iterations',
                                                content=u'Iterations',
                                                permission='view',
                                                no_link=True))
        self.actions.append(SidebarAction('view_iterations',
            content=u'View iterations',
            permission='view_iterations',
            attrs=dict(href="'%s/view_iterations' % request.application_url")))
        self.actions.append(SidebarAction('manage_iterations',
            content=u'Manage iterations',
            permission='manage_iterations',
            attrs=dict(href="'%s/manage_iterations' % request.application_url")))
        actions = self.actions.render(request)
        template =  get_renderer('penelope.core.forms:templates/project_sidebar.pt').implementation()
        return template(actions=actions,
                        request=request)

gsm = zope.component.getGlobalSiteManager()
gsm.registerAdapter(IterationSidebarRenderer, (IIterationView,), ISidebar)



import os
import sys
import transaction
import beaker
import csv

from sqlalchemy import engine_from_config
from pyramid.paster import get_appsettings, setup_logging

from penelope.core.models.dashboard import Project, Application, SVN
from penelope.core.models.dbsession import DBSession
from penelope.core.models import Base


beaker.cache.cache_regions.update(dict(calculate_matrix={}))


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri> <csv_import_file>\n'
          '(example: "%s development.ini import_svn.csv")' % (cmd, cmd)) 
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) != 3:
        usage(argv)
    config_uri = argv[1]
    svn_csv = argv[2]
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, name='dashboard')
    engine = engine_from_config(settings, 'sa.dashboard.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine
    Base.metadata.create_all()

    with transaction.manager:
        session = DBSession()
        with open(svn_csv, 'rb') as csvfile:
            reader = csv.reader(csvfile)
            for project_id, svn_repo in reader:
                svn_repo = svn_repo.strip()
                project_id = project_id.strip()
                project = session.query(Project).get(project_id)
                app = Application(name='SVN',
                                  project_id=project_id,
                                  application_type=SVN,
                                  svn_name=svn_repo)
                project.add_application(app)
                print 'Creating svn app [%s] for project [%s]' % (svn_repo, project_id)



# -*- coding: utf-8 -*-

from copy import deepcopy

from pyramid import httpexceptions as exc

from js.jqgrid import jqgrid, jqgrid_i18n_en
from fa.bootstrap import actions

from penelope.core.forms import ModelView
from penelope.core.forms import workflow
from penelope.core.forms.fast_ticketing import FastTicketing
from penelope.core.lib.helpers import ticket_url, unicodelower
from penelope.core.models import dashboard, tp, DBSession


def configurate(config):
    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='view',
                                  name='',
                                  attr='show',
                                  renderer='fa.bootstrap:templates/admin/show.pt',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='POST',
                                  permission='delete',
                                  name='delete',
                                  attr='delete',
                                  renderer='fa.bootstrap:templates/admin/edit.pt',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  renderer='fa.bootstrap:templates/admin/listing.pt',
                                  context='pyramid_formalchemy.resources.ModelListing',
                                  attr='listing',
                                  request_method='GET',
                                  permission='view',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestListingView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='workflow',
                                  name='goto_state',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=workflow.goto_state)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='estimations',
                                  name='estimations.json',
                                  renderer='json',
                                  attr='get_estimations',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='estimations',
                                  name='estimations.html',
                                  attr='get_estimations_html',
                                  renderer='penelope.core.forms:templates/estimations_simple.pt',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='POST',
                                  permission='estimations',
                                  name='estimations.json',
                                  renderer='json',
                                  attr='put_estimations',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='estimations',
                                  name='estimations',
                                  attr='estimations',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  renderer='penelope.core.forms:templates/estimations.pt',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='view',
                                  name='tickets',
                                  attr='tickets',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  renderer='penelope.core.forms:templates/customer_request_tickets.pt',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='edit',
                                  name='migrate',
                                  attr='migrate',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  renderer='penelope.core.forms:templates/customer_request_migrate.pt',
                                  view=CustomerRequestModelView)

    config.formalchemy_model_view('admin',
                                  request_method='POST',
                                  permission='edit',
                                  name='migrate',
                                  attr='do_migrate',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=CustomerRequestModelView)

    #custom view for adding tickets to a customerrequest
    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='new',
                                  name='fastticketing',
                                  attr='render',
                                  renderer='penelope.core.forms:templates/fast_ticketing.pt',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=FastTicketing)
    #custom view for adding tickets to a customerrequest
    config.formalchemy_model_view('admin',
                                  request_method='POST',
                                  permission='new',
                                  name='fastticketing',
                                  attr='render',
                                  renderer='penelope.core.forms:templates/fast_ticketing.pt',
                                  model='penelope.core.models.dashboard.CustomerRequest',
                                  view=FastTicketing)


customer_request_tabs = actions.TabsActions(actions.TabAction("show",
                                                              content="View",
                                                              permission='view',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")),
                                            actions.TabAction("tickets",
                                                              content="Tickets",
                                                              permission='view',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'tickets')")),
                                            actions.TabAction("estimations",
                                                              content="Estimations",
                                                              permission='estimations',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'estimations')")),
                                            actions.TabAction("migrate",
                                                              content="Migrate",
                                                              permission='edit',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'migrate')")),
                                            )

customer_request_tabs_without_estimations = actions.TabsActions(actions.TabAction("show",
                                                              content="View",
                                                              permission='view',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")),
                                            actions.TabAction("tickets",
                                                              content="Tickets",
                                                              permission='view',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'tickets')")),
                                            actions.TabAction("migrate",
                                                              content="Migrate",
                                                              permission='edit',
                                                              attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'migrate')")),
                                            )

add_ticket = actions.UIButton(id='add_ticket',
                              content='Add ticket',
                              permission='view',
                              _class='btn btn-success',
                              attrs=dict(href="request.model_instance.add_ticket_url(request)"))

add_fast_ticketing = actions.UIButton(id='add_fast_ticketing',
                              content='Fast add tickets',
                              permission='new',
                              _class='btn btn-success',
                              attrs=dict(href="request.fa_url('CustomerRequest', request.model_id, 'fastticketing')"))


class CustomerRequestListingView(ModelView):
    defaults_actions = deepcopy(actions.defaults_actions)
    defaults_actions.update(listing_buttons = actions.Actions())


class CustomerRequestModelView(ModelView):
    actions_categories = ('buttons', 'tabs')
    defaults_actions = deepcopy(actions.defaults_actions)

    def __init__(self, *args, **kwargs):
        super(CustomerRequestModelView, self).__init__(*args, **kwargs)
        self.defaults_actions.update(show_buttons=actions.Actions(actions.edit))
        if self.context.get_instance().filler:
            cr_actions = deepcopy(customer_request_tabs_without_estimations)
        else:
            cr_actions = deepcopy(customer_request_tabs)
        wf = workflow.change_workflow(self.context)
        if wf:
            wf_actions = cr_actions + [wf]
        self.defaults_actions.update(show_tabs = wf_actions)
        self.defaults_actions.update(estimations_tabs = cr_actions)
        self.defaults_actions['show_buttons'].append(add_ticket)
        self.defaults_actions['show_buttons'].append(add_fast_ticketing)

    def delete(self):
        cr = self.context.get_instance()
        if len(cr.get_tickets(self.request)):
            request = self.request
            request.add_message(u'Customer request has tickets. Please remove them first.', type='danger')
            raise exc.HTTPFound(location=request.fa_url('CustomerRequest', cr.id))
        else:
            return self.force_delete()

    @actions.action('estimations')
    def estimations(self):
        """Use http://trirand.com/blog grid widget """
        jqgrid_i18n_en.need()
        jqgrid.need()
        return self.render()

    def get_estimations_html(self):
        items = self.request.model_instance.estimations
        estimations =[ [item.person_type,item.days] for item in items]
        return self.render(estimations=estimations)

    @actions.action('show')
    def tickets(self):
        """Use iframe to render tickets """
        tracs = list(self.request.model_instance.project.tracs)
        if not tracs:
            return None
        trac_url = tracs[0].application_uri(self.request)
        customer_request_id = self.request.model_instance.id
        report_url = '%s/query?customerrequest=%s&iframe=true' % (trac_url, customer_request_id)
        return self.render(report_url=report_url)

    @actions.action('show')
    def migrate(self):
        """Massive time entry migration form"""
        context = self.context.get_instance()
        project = context.project
        opts = {}
        opts['crs'] = sorted([cr for cr in project.customer_requests if cr.id != context.id], key=unicodelower)
        opts['current_customer_request'] = context
        opts['back_url'] = '%s/admin/CustomerRequest/%s' % (self.request.application_url, context.id)
        opts['tes'] = context.time_entries
        opts['ticket_url'] = ticket_url
        return self.render(**opts)

    def do_migrate(self):
        from penelope.core.models.tickets import ticket_store

        context = self.context.get_instance()
        te_ids = self.request.POST.getall('te')
        new_cr = self.request.POST.get('new_cr')
        tes = DBSession().query(tp.TimeEntry).filter(tp.TimeEntry.id.in_(te_ids))
        if self.request.POST.get('submit') == u'move_time_entries_and_tickets':
            update_tickets = True
        else:
            update_tickets = False

        n_te = 0
        n_ticket = 0

        tickets = set()
        for te in tes:
            te.customer_request_id = new_cr
            tickets.add(te.ticket)
            n_te += 1

        if update_tickets:
            author = self.request.authenticated_user.email
            for ticket in tickets:
                t = ticket_store.get_raw_ticket(context.project_id, ticket)
                t['customerrequest'] = new_cr
                t.save_changes(comment=u'Massive move operation from Penelope.', author=author)
                n_ticket += 1
            self.request.add_message('%d ticket updated.' % n_ticket)

        self.request.add_message('%d time entries moved.' % n_te)
        raise exc.HTTPFound(location=self.request.fa_url('CustomerRequest', context.id, 'migrate'))

    def put_estimations(self):
        self.request.model_class = dashboard.Estimation
        self.request.model_name = 'Estimation'
        customer_request_id = self.request.model_instance.id
        response = {'success': True}

        self.data = self.request.POST
        obj_id = self.data.pop('id')
        if obj_id == u'_empty':
            obj_id = ''
        else:
            self.request.model_instance = self.session.query(dashboard.Estimation).get(obj_id)

        action = self.data.pop('oper')

        def process(session=None):
            self.data = dict([('Estimation-%s-%s' % (obj_id, k),v) for k,v in self.data.items()])
            self.data['Estimation-%s-customer_request_id' % obj_id] = customer_request_id
            if session:
                self.fs = self.fs.bind(data=self.data, session=self.session, request=self.request)
            else:
                self.fs = self.fs.bind(data=self.data, request=self.request)
            if self.validate(self.fs):
                self.fs.sync()
                return {'success': True}
            else:
                return {'success': False, 'message': 'Validation error'}

        def add():
            self.fs = self.get_fieldset()
            return process(session=self.session)

        def edit():
            self.fs = self.get_fieldset(id=obj_id)
            msg = process()
            self.sync(self.fs, obj_id)
            return msg

        def delete():
            record = self.request.model_instance
            if record:
                self.session.delete(record)

        if action == 'add': response = add()
        elif action == 'edit': response = edit()
        elif action == 'del': delete()

        self.session.flush()

        return response

    def get_estimations(self):
        """
        Return json for the grid:

          {"records":"13",
           "rows":[
             {"id":"1","cell":["1","2007-10-01","Client 1","100.00","20.00","120.00","note 1"]}],
          }
        """
        items = self.request.model_instance.estimations
        total_days = sum([a.days for a in items])
        len_items = len(items)

        resp = {'records': len_items,
                'userdata':{"days":total_days,
                            "person_type":"Totals:"},
                'rows':[]}

        for item in items:
            resp['rows'].append({'id': str(item.id), 'cell':[item.person_type,
                                                             item.days]})
        return resp




# -*- coding: utf-8 -*-
from zope.interface import implements
from copy import deepcopy

from pyramid import httpexceptions as exc
from pyramid_formalchemy import resources
from fa.bootstrap import actions
from penelope.core.interfaces import IManageView
from penelope.core.forms import ModelView
from penelope.core.forms.renderers import ProjectRelationRenderer
from penelope.core.notifications import notify_user_with_welcoming_mail
from penelope.core.models import dashboard


def configurate(config):
    #custom view for user
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='',
        attr='show',
        renderer='penelope.core.forms:templates/user.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        renderer='penelope.core.forms:templates/user_listing.pt',
        attr='datatable',
        context=ModelListing,
        request_method='GET',
        permission='listing',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    #custom view for tokens section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='edit',
        name='user_tokens',
        attr='user_tokens',
        renderer='penelope.core.forms:templates/user_tokens.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='costs',
        name='user_costs',
        attr='user_costs',
        renderer='penelope.core.forms:templates/user_costs.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='edit',
        name='send_user_invitation',
        attr='send_user_invitation',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    #custom view for adding a customer request to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='costs',
        name='add_cost',
        attr='add_cost',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='costs',
        name='add_cost',
        attr='add_cost',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)


send_user_invitation = actions.UIButton(id='send_user_invitation',
    content='Send user invitation',
    permission='edit',
    _class='btn btn-success',
    attrs=dict(href="request.fa_url('User', request.model_id, 'send_user_invitation')"))


user_tabs = actions.Actions(
        actions.TabAction("show",
            content="View",
            permission='view',
            attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")),
        actions.TabAction("user_tokens",
            content="User tokens",
            permission='edit',
            attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'user_tokens')")),
        actions.TabAction("costs",
            content="User costs",
            permission='costs',
            attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'user_costs')")),
    )

add_costs = actions.Actions(
        actions.UIButton(id='add_cost',
            content='Add user cost',
            permission='costs',
            _class='btn btn-success',
            attrs=dict(href="request.fa_url('User', request.model_id, 'add_cost')")),
        )

class ModelListing(resources.ModelListing):
    implements(IManageView)


class UserModelView(ModelView):
    actions_categories = ('buttons', 'tabs')

    defaults_actions = deepcopy(actions.defaults_actions)
    defaults_actions['show_buttons'].append(send_user_invitation)
    defaults_actions.update(show_tabs=user_tabs)
    defaults_actions.update(user_costs_tabs=user_tabs)
    defaults_actions.update(user_costs_buttons=add_costs)

    @actions.action('show')
    def user_tokens(self, *args, **kwargs):
        context = self.context.get_instance()
        return self.render(user=context)

    @actions.action('user_costs')
    def user_costs(self, *args, **kwargs):
        context = self.context.get_instance()
        page = self.get_page(collection=context.costs)
        pager = page.pager(**self.pager_args)
        return self.render(pager=pager, items=page)

    @actions.action('user_costs')
    def add_cost(self, *args, **kwargs):
        self.request.model_class = dashboard.Cost
        self.request.model_name = dashboard.Cost.__name__
        self.request.form_action = ['User', self.request.model_id, 'add_cost']
        if self.request.method == 'POST':
            self.request.POST.update({'next': self.request.fa_url('User', self.request.model_id, 'user_costs')})
        return self.new()

    @actions.action('listing')
    def datatable(self, **kwargs):
        result = super(UserModelView, self).datatable(**kwargs)

        fs = result['fs']
        fs.configure(pk=False, readonly=True)

        columns = ['email', 'fullname', 'roles', 'project_manager', 'active']
        self.pick_columns(fs, columns)

        fs._render_fields['project_manager']._get_renderer = lambda: ProjectRelationRenderer(fs._render_fields['project_manager'])

        return dict(result,
                    columns=columns)

    def delete(self):
        user = self.context.get_instance()
        request = self.request
        if not user.active:
            request.add_message(u'User is already inactive. You cannot remove it.', type='danger')
            raise exc.HTTPFound(location=request.fa_url('User', user.id))
        else:
            user.active = False
            request.add_message(u'User deactivated.', type='success')
            raise exc.HTTPFound(location=request.fa_url('User', user.id))

    def send_user_invitation(self):
        user = self.context.get_instance()
        notify_user_with_welcoming_mail(user.email)
        request = self.request
        request.add_message(u'Invitation email send.', type='success')
        raise exc.HTTPFound(location=request.fa_url('User', user.id))



# -*- coding: utf-8 -*-

import pkg_resources

from fanstatic import Group, Library, Resource
from js.jquery import jquery
from js.jqueryui import jqueryui, jqueryui_i18n
from js.jqueryui import overcast
from js.jquery_timepicker_addon import timepicker_it
from js.jquery_datatables import jquery_datatables_js    # jquery_datatables_css only contains a style for the demo table.
from js.bootstrap import bootstrap_js, bootstrap_responsive_css
from fa.bootstrap.fanstatic_resources import bootstrap
from js.socketio import socketio
from js.xeditable import bootstrap_editable


class ForeignLibrary(Library):
    """
    Let fanstatic get resources from an egg != the current one.
    Note that the library name must also be the package name, and must
    still be defined in the entry points.
    """
    def __init__(self, *args, **kw):
        super(ForeignLibrary, self).__init__(*args, **kw)
        self.path = pkg_resources.resource_filename(self.name, self.rootpath)



#----------------------
# Third party libraries
#----------------------


por_library = Library('por', 'static')

# mediaquery polyfills for IE
respond = Resource(por_library, 'respond/respond.src.js', minified='respond/respond.min.js')

# bricklayer for having columns in CSS
masonry = Resource(por_library, 'masonry/jquery.masonry.js', minified='masonry/jquery.masonry.min.js', depends=[jquery])

# JS templates
mustache = Resource(por_library, 'mustache/mustache.js', minified='mustache/mustache.min.js')

# spinner for ajax operations
spin = Resource(por_library, 'spin/spin.js')

jsonrpc = Resource(por_library, 'jquery.jsonrpc/jquery.jsonrpc.js')

jquery_pivot_css = Resource(por_library, 'jquery.pivot/jquery_pivot.css')
jquery_pivot = Resource(por_library, 'jquery.pivot/jquery_pivot.js', depends=[jquery, jquery_pivot_css])
jquery_pivot_costs = Resource(por_library, 'jquery.pivot/jquery_pivot_costs.js', depends=[jquery, jquery_pivot_css])


#-------
# Deform
#-------

deform_library = ForeignLibrary('deform', 'static')
deform_bootstrap_library = ForeignLibrary('deform_bootstrap', 'static')

deform_js = Resource(deform_library, 'scripts/deform.js', depends=[jquery])
deform_bootstrap = Group([
        Resource(deform_bootstrap_library, 'chosen_bootstrap.css'),
        Resource(deform_bootstrap_library, 'deform_bootstrap.js', depends=[deform_js]),
        Resource(por_library, 'jquery_chosen/chosen.min.css'),
        Resource(por_library, 'jquery_chosen/chosen.jquery.min.js', depends=[jquery]),
        Resource(deform_library, 'scripts/jquery.form.js', depends=[jquery]),
        ])


#----------
# Our stuff
#----------

spark = Resource(por_library, 'por_home/js/jquery.sparkline.min.js', depends=[jquery,])
dashboard_home = Group([
                        Resource(por_library, 'por_home/js/home.js', depends=[spark]),
                        Resource(por_library, 'por_home/css/home.css'),
                        Resource(por_library, 'por_home/js/outstanding_tickets.js', depends=[jquery, mustache, spin]),
                        Resource(por_library, 'por_home/css/penelope.min.css'),
                        ])

font_awesome = Resource(por_library, 'por_home/css/font-awesome.css', minified='por_home/css/font-awesome.min.css', depends=[bootstrap_responsive_css])
dashboard_js = Resource(por_library, 'js/dashboard.js', depends=[bootstrap_js, jqueryui, timepicker_it])
dashboard_css = Resource(por_library, 'css/dashboard.css', depends=[bootstrap_responsive_css, respond, font_awesome])

project_filter = Resource(por_library, 'por_report/select_project_filter.js')

subnav = Group([
                Resource(por_library, 'por_subnav/subnav.js', depends=[jquery]),
                Resource(por_library, 'por_subnav/subnav.css'),
               ])

all_entries = Group([
    Resource(por_library, 'por_report/all_entries.js', depends=[jquery, timepicker_it]),
    Resource(por_library, 'por_report/all_entries.css'),
    ])

report_all_entries = Group([
    jquery_pivot,
    all_entries
    ])

report_costs = Group([
    jquery_pivot_costs,
    all_entries
    ])

saved_queries = Resource(por_library, 'por_report/saved_queries.js', depends=[jquery])

report_te_state_change = Group([
                                Resource(por_library, 'por_report/te_state_change.js', depends=[jquery, timepicker_it]),
                                Resource(por_library, 'por_report/te_state_change.css'),
                                ])

add_entry_from_ticket = Group([
                                Resource(por_library, 'por_add_entry/add_entry_from_ticket.js', depends=[jquery, mustache, timepicker_it, jsonrpc]),
                                Resource(por_library, 'por_add_entry/add_entry_from_ticket.css'),
                                ])

wizard = Group([ Resource(por_library, 'por_wizard/wizard.css') ])
fastticketing = Group([ Resource(por_library, 'por_fastticketing/fastticketing.css') ])

add_entry = Group([
                    Resource(por_library, 'por_add_entry/add_entry.js', depends=[jquery]),
                    Resource(por_library, 'por_add_entry/add_entry.css'),
                    Resource(por_library, 'por_add_entry/autocomplete.js', depends=[jquery]),
                    spin,
                   ])

backlog = Group([
                    Resource(por_library, 'por_backlog/backlog.js', depends=[jquery]),
                    Resource(por_library, 'por_backlog/backlog.css'),
                    ])

datatables = Group([
                    Resource(por_library, 'por_datatables/js/sorting.js'),
                    Resource(por_library, 'por_datatables/js/paging.js', depends=[jquery_datatables_js]),
                    Resource(por_library, 'por_datatables/css/paging.css'),
                   ])

angular = Resource(por_library, 'por_kanban/js/angular.min.js', depends=[jqueryui])
angular_animation = Resource(por_library, 'por_kanban/js/angular-animate.min.js', depends=[angular])

angular_dd = Resource(por_library,
                      'por_kanban/js/angular-dragdrop.js',
                      minified='por_kanban/js/angular-dragdrop.min.js',
                      depends=[jqueryui, angular])

angular_bootstrap_ui = Resource(por_library,
                      'por_kanban/js/ui-bootstrap-tpls-0.5.0.min.js',
                      depends=[bootstrap_js, angular])

angular_sortable = Resource(por_library,
                      'por_kanban/js/sortable.js',
                      depends=[jqueryui, angular])

js_md5 = Resource(por_library,
                  'por_kanban/js/md5.js',
                  minified='por_kanban/js/md5.min.js',)

pnotify = Group([ 
                   Resource(por_library, 'por_kanban/css/jquery.pnotify.default.css'),
                   Resource(por_library, 'por_kanban/css/jquery.pnotify.default.icons.css'),
                   Resource(por_library, 'por_kanban/js/jquery.pnotify.js',
                            minified='por_kanban/js/jquery.pnotify.min.js',
                            depends=[jqueryui])
                   ])

kanban = Group([
                Resource(por_library, 'por_kanban/js/kanban.js', depends=[bootstrap_js, angular_sortable, socketio, angular_animation, bootstrap_editable, js_md5, angular_bootstrap_ui, pnotify]),
                Resource(por_library, 'por_kanban/css/kanban.css', depends=[bootstrap]),
    ])

dashboard = Group([jquery, deform_bootstrap, jqueryui, jqueryui_i18n, overcast, bootstrap, dashboard_js, dashboard_css, subnav])
